__all__ = ["raise_not_implemented_error"]

import functools


def raise_not_implemented_error(method):
    @functools.wraps(method)
    def wrapper(self, *args, **kwargs):
        raise NotImplementedError(
            f"The linear operator '{self.__class__.__name__}' provides no "
            f"implementation for '{method.__name__}'"
        )

    return wrapper



import typing

from petsc4py import PETSc
from slepc4py import SLEPc

from .linear_operator import LinearOperator


class ProductLinearOperator(LinearOperator):
    r"""
    Linear operator of the form

    .. math::

        L = L_{r}L_{r-1}\ldots L_{2}L_{1}

    where :math:`L_i` are instances of the :class:`.LinearOperator` class.
    For example, by setting :code:`linops = (L2, L1)` and
    :code:`linops_actions = (L2.solve, L1.apply_hermitian_transpose)`
    we define a linear operator

    .. math::

        L = L_2^{-1}L_1^*.


    :param linops: list of linear operators (see example above).
    :type linops: List[LinearOperator]
    :param linops_actions: list of actions (see example above). Must be one of
        :meth:`.LinearOperator.apply`, :meth:`.LinearOperator.solve`,
        :meth:`.LinearOperator.apply_hermitian_transpose` or
        :meth:`.LinearOperator.solve_hermitian_transpose`
    :type linops_actions: List[Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec]]
    :param nblocks: number of blocks (if the operator has block structure)
    :type nblocks: Optional[Union[int, None]], default is None
    """

    def __init__(
        self: "ProductLinearOperator",
        linops: typing.List[LinearOperator],
        linops_actions: typing.List[
            typing.Callable[[PETSc.Vec, typing.Optional[PETSc.Vec]], PETSc.Vec]
        ],
        nblocks: typing.Optional[int] = None,
    ) -> None:
        linops.reverse()
        linops_actions.reverse()
        if len(linops) == 1:
            raise ValueError(
                f"ProductLinearOperator is the product of at least two "
                f"linear operators. Only one was provided at initialization."
            )
        if len(linops) != len(linops_actions):
            raise ValueError(
                f"len(linops) must be equal to len(linops_actions)."
            )

        self.linops = linops
        self.actions = linops_actions
        self.actions_hermitian_transpose = []
        self.actions_mat = []
        self.actions_hermitian_transpose_mat = []
        self.names = []
        for i, linop in enumerate(self.linops):
            name = "L%03d" % i
            self.names.append(name)
            setattr(self, name, linop)
            if self.actions[i] == linop.apply:
                self.actions_hermitian_transpose.append(
                    linop.apply_hermitian_transpose
                )
                self.actions_mat.append(linop.apply_mat)
                self.actions_hermitian_transpose_mat.append(
                    linop.apply_hermitian_transpose_mat
                )
            elif self.actions[i] == linop.apply_hermitian_transpose:
                self.actions_hermitian_transpose.append(linop.apply)
                self.actions_mat.append(linop.apply_hermitian_transpose_mat)
                self.actions_hermitian_transpose_mat.append(linop.apply_mat)
            elif self.actions[i] == linop.solve:
                self.actions_hermitian_transpose.append(
                    linop.solve_hermitian_transpose
                )
                self.actions_mat.append(linop.solve_mat)
                self.actions_hermitian_transpose_mat.append(
                    linop.solve_hermitian_transpose_mat
                )
            elif self.actions[i] == linop.solve_hermitian_transpose:
                self.actions_hermitian_transpose.append(linop.solve)
                self.actions_mat.append(linop.solve_hermitian_transpose_mat)
                self.actions_hermitian_transpose_mat.append(linop.solve_mat)
            else:
                raise ValueError(
                    f"Action for the {i}th linear operator in the list must be "
                    f"one of apply, solve, apply_hermitian_transpose or "
                    f"solve_hermitian_transpose."
                )

        self.actions_hermitian_transpose.reverse()
        self.actions_hermitian_transpose_mat.reverse()

        self.nlops = len(self.names)
        self.create_intermediate_vectors()
        L1 = getattr(self, self.names[0])
        Ln = getattr(self, self.names[-1])
        dimr = (
            Ln.get_dimensions()[0]
            if self.actions[-1] == Ln.apply or self.actions[-1] == Ln.solve
            else Ln.get_dimensions()[-1]
        )
        dimc = (
            L1.get_dimensions()[-1]
            if self.actions[0] == L1.apply or self.actions[0] == L1.solve
            else Ln.get_dimensions()[0]
        )
        dims = (dimr, dimc)
        super().__init__(
            linops[0].get_comm(), "ProductLinearOperator", dims, nblocks
        )

    def create_intermediate_vectors(self: "ProductLinearOperator") -> None:
        self.intermediate_vecs = []
        for j in range(self.nlops - 1):
            L = getattr(self, self.names[j])
            if self.actions[j] == L.apply or self.actions[j] == L.solve:
                self.intermediate_vecs.append(L.create_left_vector())
            else:
                self.intermediate_vecs.append(L.create_right_vector())
        self.intermediate_vecs_hermitian_transpose = []

        self.names.reverse()
        self.actions.reverse()
        for j in range(self.nlops - 1):
            L = getattr(self, self.names[j])
            if self.actions[j] == L.apply or self.actions[j] == L.solve:
                self.intermediate_vecs_hermitian_transpose.append(
                    L.create_right_vector()
                )
            else:
                self.intermediate_vecs_hermitian_transpose.append(
                    L.create_left_vector()
                )
        self.names.reverse()
        self.actions.reverse()

    def create_intermediate_bvs(
        self: "ProductLinearOperator", m: int
    ) -> typing.List[SLEPc.BV]:
        r"""
        Create list of matrices to store :math:`L_i Z`, for any matrix
        Z with :math:`m` columns.

        .. attention::

            It is the user's responsibility to destroy this object
            when no longer needed.

        :param m: number of columns of :math:`Z`
        :type m: int

        :rtype: List[SLEPc.BV]
        """
        intermediate_bvs = []
        for j in range(self.nlops - 1):
            L = getattr(self, self.names[j])
            X = SLEPc.BV().create(self.get_comm())
            sz = (
                L.get_dimensions()[0]
                if self.actions[j] == L.apply or self.actions[j] == L.solve
                else L.get_dimensions()[-1]
            )
            X.setSizes(sz, m)
            X.setType("mat")
            intermediate_bvs.append(X)
        return intermediate_bvs

    def create_intermediate_bvs_hermitian_transpose(
        self: "ProductLinearOperator", m: int
    ) -> typing.List[SLEPc.BV]:
        r"""
        Create list of matrices to store :math:`L_i^* Z`, for any matrix
        Z with :math:`m` columns.

        .. attention::

            It is the user's responsibility to destroy this object
            when no longer needed.

        :param m: number of columns of :math:`Z`
        :type m: int

        :rtype: List[SLEPc.BV]
        """
        self.names.reverse()
        self.actions.reverse()
        intermediate_bvs = []
        for j in range(self.nlops - 1):
            L = getattr(self, self.names[j])
            X = SLEPc.BV().create(self.get_comm())
            sz = (
                L.get_dimensions()[-1]
                if self.actions[j] == L.apply or self.actions[j] == L.solve
                else L.get_dimensions()[0]
            )
            X.setSizes(sz, m)
            X.setType("mat")
            intermediate_bvs.append(X)
        self.names.reverse()
        self.actions.reverse()
        return intermediate_bvs

    def apply(self, x, y=None):
        y = self.create_left_vector() if y == None else y
        for j in range(self.nlops):
            if j == 0:
                yj = self.intermediate_vecs[j]
                yj = self.actions[j](x, yj)
                self.intermediate_vecs[j] = yj
            elif j > 0 and j < self.nlops - 1:
                yj = self.intermediate_vecs[j]
                yjm1 = self.intermediate_vecs[j - 1]
                yj = self.actions[j](yjm1, yj)
                self.intermediate_vecs[j] = yj
            else:
                yjm1 = self.intermediate_vecs[j - 1]
                y = self.actions[j](yjm1, y)
        return y

    def apply_hermitian_transpose(self, x, y=None):
        y = self.create_right_vector() if y == None else y
        for j in range(self.nlops):
            if j == 0:
                yj = self.intermediate_vecs_hermitian_transpose[j]
                yj = self.actions_hermitian_transpose[j](x, yj)
                self.intermediate_vecs_hermitian_transpose[j] = yj
            elif j > 0 and j < self.nlops - 1:
                yj = self.intermediate_vecs_hermitian_transpose[j]
                yjm1 = self.intermediate_vecs_hermitian_transpose[j - 1]
                yj = self.actions_hermitian_transpose[j](yjm1, yj)
                self.intermediate_vecs_hermitian_transpose[j] = yj
            else:
                yjm1 = self.intermediate_vecs_hermitian_transpose[j - 1]
                y = self.actions_hermitian_transpose[j](yjm1, y)
        return y

    def apply_mat(self, X, Y=None, Z=None):
        Y = self.create_left_bv(X.getSizes()[-1]) if Y == None else Y
        destroy = False
        if Z == None:
            Z = self.create_intermediate_bvs(X.getSizes()[-1])
            destroy = True
        for j in range(self.nlops):
            if j == 0:
                Z[j] = self.actions_mat[j](X, Z[j])
            elif j > 0 and j < self.nlops - 1:
                Z[j] = self.actions_mat[j](Z[j - 1], Z[j])
            else:
                Y = self.actions_mat[j](Z[j - 1], Y)
        if destroy:
            for Zj in Z:
                Zj.destroy()
        return Y

    def apply_hermitian_transpose_mat(self, X, Y=None, Z=None):
        Y = self.create_right_bv(X.getSizes()[-1]) if Y == None else Y
        destroy = False
        if Z == None:
            Z = self.create_intermediate_bvs_hermitian_transpose(
                X.getSizes()[-1]
            )
            destroy = True
        for j in range(self.nlops):
            if j == 0:
                Z[j] = self.actions_hermitian_transpose_mat[j](X, Z[j])
            elif j > 0 and j < self.nlops - 1:
                Z[j] = self.actions_hermitian_transpose_mat[j](Z[j - 1], Z[j])
            else:
                Y = self.actions_hermitian_transpose_mat[j](Z[j - 1], Y)
        if destroy:
            for Zj in Z:
                Zj.destroy()
        return Y

    def destroy_intermediate_vectors(self: "ProductLinearOperator"):
        for vec in self.intermediate_vecs:
            vec.destroy()
        for vec in self.intermediate_vecs_hermitian_transpose:
            vec.destroy()

    def destroy(self):
        self.destroy_intermediate_vectors()
        for name in self.names:
            getattr(self, name).destroy()



import typing

import numpy as np
import scipy as sp
from petsc4py import PETSc
from slepc4py import SLEPc

from ..utils.bv import bv_add
from .linear_operator import LinearOperator
from .low_rank import LowRankLinearOperator


class LowRankUpdatedLinearOperator(LinearOperator):
    r"""
    Class for a linear operator of the form

    .. math::

        L = A + B K C^*

    where :math:`A` is an instance of the :class:`.LinearOperator` class,
    and :math:`B`, :math:`K` and :math:`C` are low-rank (dense) matrices of
    conformal sizes. If :code:`A.solve()` is enabled, then the :code:`solve()`
    method in this class is implemented using the Woodbury matrix identity

    .. math::

        L^{-1} = A^{-1} - X D Y^*,

    where :math:`X`, :math:`D` and :math:`Y` the Woodbury factors defined as

    .. math::

        \textcolor{black}{X}
        = A^{-1}B,\quad \textcolor{black}{D} =
        K\left(I + C^* A^{-1}B K\right)^{-1},\quad
        \textcolor{black}{Y} = A^{-*}C,


    :param A: instance of the :class:`.LinearOperator` class
    :param B: tall and skinny matrix
    :type B: SLEPc.BV
    :param K: dense matrix
    :type K: numpy.ndarray
    :param C: tall and skinny matrix
    :type C: SLEPc.BV
    :param woodbury_factors: tuple :math:`(X, D, Y)` of Woodbury factors. If
        :code:`A.solve()` is enabled and the argument :code:`woodbury_factors`
        is :code:`None`, the factors :math:`X`, :math:`D` and :math:`Y` are
        computed at initialization
    :type woodbury_factors: Optional[Union[Tuple[SLEPc.BV, numpy.ndarray,
        SLEPc.BV], None]], default is None
    :param nblocks: number of blocks (if the operator has block structure)
    :type nblocks: Optional[Union[int, None]], default is None
    """

    def __init__(
        self: "LowRankUpdatedLinearOperator",
        A: LinearOperator,
        B: SLEPc.BV,
        K: np.ndarray,
        C: SLEPc.BV,
        woodbury_factors: typing.Optional[
            typing.Union[typing.Tuple[SLEPc.BV, np.ndarray, SLEPc.BV], None]
        ] = None,
        nblocks: typing.Optional[typing.Union[int, None]] = None,
    ) -> None:
        comm = A.get_comm()
        self.A = A
        self.L = LowRankLinearOperator(B, K, C, nblocks)
        self.W = (
            self.compute_woodbury_operator(nblocks)
            if woodbury_factors == None
            else LowRankLinearOperator(*woodbury_factors, nblocks)
        )
        self.create_intermediate_vectors()
        super().__init__(
            comm, "LowRankUpdatedLinearOperator", A.get_dimensions(), nblocks
        )

    def compute_woodbury_operator(
        self: "LowRankUpdatedLinearOperator",
        nblocks: typing.Union[int, None],
    ) -> LowRankLinearOperator:
        r"""
        :param nblocks: number of blocks (if the operator has block structure)
        :type nblocks: Unions[int, None]

        :return: a :class:`.LowRankLinearOperator` constructed from the
            Woodbury factors :code:`X`, :code:`D` and :code:`Y`
        :rtype: LowRankUpdatedLinearOperator
        """
        comm = self.A.get_comm()
        try:
            X = self.A.solve_mat(self.L.U)
            Y = self.A.solve_hermitian_transpose_mat(self.L.V)
            S = PETSc.Mat().createDense(
                self.L.Sigma.shape, None, self.L.Sigma, PETSc.COMM_SELF
            )
            XS = SLEPc.BV().create(comm)
            XS.setSizes(X.getSizes()[0], self.L.Sigma.shape[-1])
            XS.setType("mat")
            XS.mult(1.0, 0.0, X, S)
            M = XS.dot(self.L.V)
            Ma = M.getDenseArray()
            D = self.L.Sigma @ sp.linalg.inv(np.eye(Ma.shape[0]) + Ma)
            W = LowRankLinearOperator(X, D, Y, nblocks)
            XS.destroy()
            M.destroy()
            S.destroy()
        except:
            W = None
        return W

    def create_intermediate_vectors(
        self: "LowRankUpdatedLinearOperator",
    ) -> None:
        self.Ax = self.A.create_left_vector()
        self.ATx = self.A.create_right_vector()

    def create_intermediate_bv(
        self: "LowRankUpdatedLinearOperator", m: int
    ) -> SLEPc.BV:
        r"""
        Create matrix to store :math:`X D Y^* Z`, for any matrix
        Z with :math:`m` columns.

        .. attention::

            It is the user's responsibility to destroy this object
            when no longer needed.

        :param m: number of columns of :math:`Z`
        :type m: int

        :rtype: SLEPc.BV
        """
        X = SLEPc.BV().create(self.get_comm())
        X.setSizes(self.get_dimensions()[0], m)
        X.setType("mat")
        return X

    def create_intermediate_bv_hermitian_transpose(
        self: "LowRankUpdatedLinearOperator", m: int
    ) -> SLEPc.BV:
        r"""
        Create matrix to store :math:`Y D X^* Z`, for any matrix
        Z with :math:`m` columns.

        .. attention::

            It is the user's responsibility to destroy this object
            when no longer needed.

        :param m: number of columns of :math:`Z`
        :type m: int

        :rtype: SLEPc.BV
        """
        X = SLEPc.BV().create(self.get_comm())
        X.setSizes(self.get_dimensions()[-1], m)
        X.setType("mat")
        return X

    def apply(self, x, y=None):
        self.Ax = self.A.apply(x, self.Ax)
        y = self.L.apply(x, y)
        y.axpy(1.0, self.Ax)
        return y

    def apply_hermitian_transpose(self, x, y=None):
        self.ATx = self.A.apply_hermitian_transpose(x, self.ATx)
        y = self.L.apply_hermitian_transpose(x, y)
        y.axpy(1.0, self.ATx)
        return y

    def apply_mat(self, X, Y=None, Z=None):
        destroy = False
        if Z == None:
            destroy = True
            Z = self.create_intermediate_bv(X.getSizes()[-1])
        Z = self.A.apply_mat(X, Z)
        Y = self.L.apply_mat(X, Y)
        bv_add(1.0, Y, Z)
        Z.destroy() if destroy else None
        return Y

    def apply_hermitian_transpose_mat(self, X, Y=None, Z=None):
        destroy = False
        if Z == None:
            destroy = True
            Z = self.create_intermediate_bv_hermitian_transpose(
                X.getSizes()[-1]
            )
        Z = self.A.apply_hermitian_transpose_mat(X, Z)
        Y = self.L.apply_hermitian_transpose_mat(X, Y)
        bv_add(1.0, Y, Z)
        Z.destroy() if destroy else None
        return Y

    def solve(self, x, y=None):
        self.Ax = self.A.solve(x, self.Ax)
        y = self.W.apply(x, y)
        y.scale(-1.0)
        y.axpy(1.0, self.Ax)
        return y

    def solve_hermitian_transpose(self, x, y=None):
        self.ATx = self.A.solve_hermitian_transpose(x, self.ATx)
        y = self.W.apply_hermitian_transpose(x, y)
        y.scale(-1.0)
        y.axpy(1.0, self.ATx)
        return y

    def solve_mat(self, X, Y=None, Z=None):
        destroy = False
        if Z == None:
            destroy = True
            Z = self.create_intermediate_bv(X.getSizes()[-1])
        Z = self.A.solve_mat(X, Z)
        Y = self.W.apply_mat(X, Y)
        Y.scale(-1.0)
        bv_add(1.0, Y, Z)
        Z.destroy() if destroy else None
        return Y

    def solve_hermitian_transpose_mat(self, X, Y=None, Z=None):
        destroy = False
        if Z == None:
            destroy = True
            Z = self.create_intermediate_bv_hermitian_transpose(
                X.getSizes()[-1]
            )
        Z = self.A.solve_hermitian_transpose_mat(X, Z)
        Y = self.W.apply_hermitian_transpose_mat(X, Y)
        Y.scale(-1.0)
        bv_add(1.0, Y, Z)
        Z.destroy() if destroy else None
        return Y

    def destroy_woodbury_operator(
        self: "LowRankUpdatedLinearOperator",
    ) -> None:
        self.W.destroy() if self.W is not None else None

    def destroy_low_rank_update(self: "LowRankUpdatedLinearOperator") -> None:
        self.L.destroy()

    def destroy_intermediate_vectors(
        self: "LowRankUpdatedLinearOperator",
    ) -> None:
        self.Ax.destroy()
        self.ATx.destroy()

    def destroy(self):
        self.destroy_intermediate_vectors()
        self.destroy_woodbury_operator()
        self.destroy_low_rank_update()



import typing

from petsc4py import PETSc
from slepc4py import SLEPc

from ..utils.matrix import mat_solve_hermitian_transpose
from .linear_operator import LinearOperator


class MatrixLinearOperator(LinearOperator):
    r"""
    Class for a linear operator :math:`L = A`, where :math:`A` is a matrix.
    In general, :code:`A` can be any matrix (rectangular or square,
    invertible or non-invertible). For an invertible :math:`A`, the user
    may also provide a PETSc KSP object :code:`ksp` to act with :math:`A^{-1}`
    on vectors and matrices.

    :param A: a PETSc matrix
    :type A: PETSc.Mat
    :param ksp: a PETSc KSP object to enable the :code:`solve()`
        and :code:`solve_hermitian_transpose()` methods
    :type ksp: Optional[Union[PETSc.KSP, None]], defaults to None
    :param nblocks: number of blocks (if the linear operator has block
        structure). This must be an odd number.
    :type nblocks: Optional[Union[int, None]], default is None
    """

    def __init__(
        self: "MatrixLinearOperator",
        A: PETSc.Mat,
        ksp: typing.Optional[typing.Union[PETSc.KSP, None]] = None,
        nblocks: typing.Optional[typing.Union[int, None]] = None,
    ) -> None:
        self.A = A
        self.ksp = ksp
        super().__init__(
            A.getComm(), "MatrixLinearOperator", A.getSizes(), nblocks
        )

    def apply(self, x, y=None):
        y = self.create_left_vector() if y == None else y
        self.A.mult(x, y)
        return y

    def apply_mat(self, X, Y=None):
        Xm = X.getMat()
        if Y != None:
            Ym = Y.getMat()
            Ym = self.A.matMult(Xm, Ym)
            Y.restoreMat(Ym)
        else:
            Ym = Xm.duplicate()
            Ym = self.A.matMult(Xm, Ym)
            Y = SLEPc.BV().createFromMat(Ym)
            Y.setType("mat")
            Ym.destroy()
        X.restoreMat(Xm)
        return Y

    def hermitian_transpose(self):
        self.A.hermitianTranspose()

    def apply_hermitian_transpose(self, x, y=None):
        y = self.create_right_vector() if y == None else y
        self.A.multHermitian(x, y)
        return y

    def apply_hermitian_transpose_mat(self, X, Y=None):
        self.A.hermitianTranspose()
        Xm = X.getMat()
        if Y != None:
            Ym = Y.getMat()
            Ym = self.A.matMult(Xm, Ym)
            Y.restoreMat(Ym)
        else:
            Ym = Xm.duplicate()
            Ym = self.A.matMult(Xm, Ym)
            Y = SLEPc.BV().createFromMat(Ym)
            Y.setType("mat")
            Ym.destroy()
        X.restoreMat(Xm)
        self.A.hermitianTranspose()
        return Y

    def solve(self, x, y=None):
        if self.ksp != None:
            y = self.create_left_vector() if y == None else y
            self.ksp.solve(x, y)
            return y
        else:
            raise Exception(
                f"Error from {self.get_name()}.solve(): "
                f"Please provide a PETSc KSP object at initialization to use "
                f"the solve() method."
            )

    def solve_mat(self, X, Y=None):
        if self.ksp != None:
            Xm = X.getMat()
            if Y != None:
                Ym = Y.getMat()
                self.ksp.matSolve(Xm, Ym)
                Y.restoreMat(Ym)
            else:
                Ym = Xm.duplicate()
                self.ksp.matSolve(Xm, Ym)
                Y = SLEPc.BV().createFromMat(Ym)
                Y.setType("mat")
                Ym.destroy()
            X.restoreMat(Xm)
            return Y
        else:
            raise Exception(
                f"Error from {self.get_name()}.solve(): "
                f"Please provide a PETSc KSP object at initialization to use "
                f"the solve() method."
            )

    def solve_hermitian_transpose(self, x, y=None):
        if self.ksp != None:
            y = self.create_right_vector() if y == None else y
            x.conjugate()
            self.ksp.solveTranspose(x, y)
            x.conjugate()
            y.conjugate()
            return y
        else:
            raise Exception(
                f"Error from {self.get_name()}.solve_hermitian_transpose(). "
                f"Please provide a PETSc KSP object at initialization to use "
                f"the solve_hermitian_transpose() method."
            )

    def solve_hermitian_transpose_mat(self, X, Y=None):
        if self.ksp != None:
            Xm = X.getMat()
            if Y != None:
                Ym = Y.getMat()
                Ym = mat_solve_hermitian_transpose(self.ksp, Xm, Ym)
                Y.restoreMat(Ym)
            else:
                Ym = Xm.duplicate()
                Ym = mat_solve_hermitian_transpose(self.ksp, Xm, Ym)
                Y = SLEPc.BV().createFromMat(Ym)
                Y.setType("mat")
                Ym.destroy()
            X.restoreMat(Xm)
            return Y
        else:
            raise Exception(
                f"Error from {self.get_name()}.solve(): "
                f"Please provide a PETSc KSP object at initialization to use "
                f"the solve() method."
            )

    def destroy_matrix(self: "MatrixLinearOperator"):
        self.A.destroy()

    def destroy_ksp(self: "MatrixLinearOperator"):
        self.ksp.destroy() if self.ksp is not None else None

    def destroy(self):
        self.destroy_matrix()
        self.destroy_ksp()



import typing

import scipy as sp
from petsc4py import PETSc
from slepc4py import SLEPc

from ..utils.matrix import create_AIJ_identity
from .linear_operator import LinearOperator
from .low_rank import LowRankLinearOperator
from .low_rank_updated import LowRankUpdatedLinearOperator
from .matrix import MatrixLinearOperator


class ProjectionLinearOperator(LinearOperator):
    r"""
    Class for a linear operator of the form

    .. math::

        L = \begin{cases}
        \Phi \left(\Psi^*\Phi\right)^{-1} \Psi^{*} & \text{if } \texttt{complement=False} \\ 
        I - \Phi \left(\Psi^*\Phi\right)^{-1} \Psi^{*} & \text{otherwise}
        \end{cases}

    where :math:`\Phi` and :math:`\Psi` are tall and skinny matrices of size
    :math:`N\times r` stored as SLEPc BVs. (It is easy to check that :math:`L` 
    is a projection since :math:`L^2 = L`.)
    
    :param Phi: tall and skinny matrix
    :type Phi: SLEPc.BV
    :param Psi: tall and skinny matrix
    :type Psi: SLEPc.BV
    :param complement: see definition of :math:`L`
    :type complement: Optional[bool], default if False
    :param nblocks: number of blocks (if the operator has block structure)
    :type nblocks: Optional[Union[int, None]], default is None
    """

    def __init__(
        self: "ProjectionLinearOperator",
        Phi: SLEPc.BV,
        Psi: SLEPc.BV,
        complement: typing.Optional[bool] = False,
        nblocks: typing.Optional[int] = None,
    ) -> None:
        comm = Phi.getComm()
        ncolsPhi = Phi.getSizes()[-1]
        ncolsPsi = Psi.getSizes()[-1]
        if ncolsPhi != ncolsPsi:
            raise ValueError(
                "Phi and Psi should have the same number of columns."
            )
        dimensions = (Phi.getSizes()[0], Psi.getSizes()[0])
        Smat = Phi.dot(Psi)
        Sig = sp.linalg.inv(Smat.getDenseArray())
        Smat.destroy()

        self.complement = complement
        if self.complement:
            Id = create_AIJ_identity(comm, dimensions)
            self.Idop = MatrixLinearOperator(Id, None, nblocks)
            self.L = LowRankUpdatedLinearOperator(
                self.Idop, Phi, -Sig, Psi, None, nblocks
            )
        else:
            self.L = LowRankLinearOperator(Phi, Sig, Psi, nblocks)

        super().__init__(comm, "ProjectionLinearOperator", dimensions, nblocks)

    def apply(self, x, y=None):
        return self.L.apply(x, y)

    def apply_hermitian_transpose(self, x, y=None):
        return self.L.apply_hermitian_transpose(x, y)

    def apply_mat(self, X, Y=None):
        return self.L.apply_mat(X, Y)

    def apply_hermitian_transpose_mat(self, X, Y=None):
        return self.L.apply_hermitian_transpose_mat(X, Y)

    def destroy(self):
        self.L.destroy()
        if self.complement:
            self.Idop.destroy()



from .bv import *
from .comms import *
from .errors import *
from .io import *
from .ksp import *
from .matrix import *
from .miscellaneous import *
from .random import *
from .vector import *

del bv, comms, errors, io, ksp, matrix, miscellaneous, random, vector



import typing

import numpy as np
from petsc4py import PETSc
from slepc4py import SLEPc

from .linear_operator import LinearOperator


class LowRankLinearOperator(LinearOperator):
    r"""
    Class for a linear operator of the form

    .. math::

        L = U \Sigma V^*,

    where :math:`U`, :math:`\Sigma` and :math:`V` are matrices of
    conformal sizes (and :math:`\Sigma` is not necessarily diagonal).

    :param U: a tall and skinny matrix
    :type U: SLEPc.BV
    :param Sigma: 2D numpy array
    :type Sigma: numpy.ndarray
    :param V: a tall and skinny matrix
    :type V: SLEPc.BV
    :param nblocks: number of blocks (if the operator has block structure)
    :type nblocks: Optional[Union[int, None]], default is None
    """

    def __init__(
        self: "LowRankLinearOperator",
        U: SLEPc.BV,
        Sigma: np.ndarray,
        V: SLEPc.BV,
        nblocks: typing.Optional[typing.Union[int, None]] = None,
    ) -> None:
        self.U = U
        self.Sigma = Sigma
        self.V = V
        dimensions = (U.getSizes()[0], V.getSizes()[0])
        super().__init__(
            U.getComm(), "LowRankLinearOperator", dimensions, nblocks
        )

    def apply(self, x, y=None):
        y = self.create_left_vector() if y == None else y
        q = self.Sigma @ self.V.dotVec(x)
        self.U.multVec(1.0, 0.0, y, q)
        return y

    def apply_hermitian_transpose(self, x, y=None):
        y = self.create_right_vector() if y == None else y
        q = self.Sigma.conj().T @ self.U.dotVec(x)
        self.V.multVec(1.0, 0.0, y, q)
        return y

    def apply_mat(self, X, Y=None):
        M = X.dot(self.V)
        L = self.Sigma @ M.getDenseArray()
        Lm = PETSc.Mat().createDense(L.shape, None, L, PETSc.COMM_SELF)
        Y = X.duplicate() if Y == None else Y
        Y.mult(1.0, 0.0, self.U, Lm)
        Lm.destroy()
        M.destroy()
        return Y

    def apply_hermitian_transpose_mat(self, X, Y=None):
        M = X.dot(self.U)
        L = self.Sigma.conj().T @ M.getDenseArray()
        Lm = PETSc.Mat().createDense(L.shape, None, L, PETSc.COMM_SELF)
        Y = X.duplicate() if Y == None else Y
        Y.mult(1.0, 0.0, self.V, Lm)
        Lm.destroy()
        M.destroy()
        return Y

    def destroy_U(self: "LowRankLinearOperator") -> None:
        self.U.destroy()

    def destroy_V(self: "LowRankLinearOperator") -> None:
        self.V.destroy()

    def destroy_Sigma(self: "LowRankLinearOperator") -> None:
        del self.Sigma

    def destroy(self):
        self.destroy_U()
        self.destroy_V()
        self.destroy_Sigma()



__all__ = [
    "arnoldi_iteration",
    "eig",
    "match_right_and_left_eigenvectors",
    "check_eig_convergence",
]

import typing

import numpy as np
import scipy as sp
from mpi4py import MPI
from petsc4py import PETSc
from slepc4py import SLEPc

from ..linear_operators import LinearOperator
from ..utils.miscellaneous import petscprint
from ..utils.random import generate_random_petsc_vector
from ..utils.vector import enforce_complex_conjugacy
from ..utils.matrix import create_dense_matrix


def arnoldi_iteration(
    L: LinearOperator,
    action: typing.Callable[
        [PETSc.Vec, typing.Optional[PETSc.Vec]], PETSc.Vec
    ],
    krylov_dim: int,
    verbose: typing.Optional[int] = 0,
) -> typing.Tuple[SLEPc.BV, np.ndarray]:
    r"""
    Perform the Arnoldi iteration algorithm to compute an 
    orthonormal basis and the corresponding Hessenberg matrix 
    for the range of the linear operator specified by
    :code:`L` and :code:`action`.
    
    :param L: instance of the :class:`.LinearOperator` class
    :type L: :class:`.LinearOperator`
    :param action: one of :meth:`.LinearOperator.apply`, 
        :meth:`.LinearOperator.apply_hermitian_transpose`, 
        :meth:`.LinearOperator.solve` or 
        :meth:`.LinearOperator.solve_hermitian_transpose`
    :type action: Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec]
    :param krylov_dim: dimension of the Arnoldi Krylov subspace
    :type krylov_dim: int
    :param verbose: 0 = no printout to terminal, 1 = print progress
    :type verbose: Optional[int], default is 0
    
    :return: tuple with an orthonormal basis for the Krylov subspace
        and the Hessenberg matrix
    :rtype: (`BV`_ with :code:`krylov_dim` columns, \
        numpy.ndarray of size :code:`krylov_dim x krylov_dim`)
    """
    comm = L.get_comm()
    sizes = (
        L.get_dimensions()[0]
        if action == L.apply or action == L.solve
        else L.get_dimensions()[1]
    )
    nblocks = L.get_nblocks()
    block_cc = L.get_block_cc_flag()
    # Initialize the Hessenberg matrix and the BV for the Krylov subspace
    Q = SLEPc.BV().create(comm=comm)
    Q.setSizes(sizes, krylov_dim + 1)
    Q.setType("mat")
    H = np.zeros((krylov_dim + 1, krylov_dim), dtype=np.complex128)
    complex = False if L.get_real_flag() else True
    q = generate_random_petsc_vector(sizes, complex)
    enforce_complex_conjugacy(comm, q, nblocks) if block_cc == True else None
    q.scale(1.0 / q.norm())
    Q.insertVec(0, q)
    # Perform the Arnoldi iterations
    v = (
        L.create_left_vector()
        if action == L.apply or action == L.solve
        else L.create_right_vector()
    )
    for k in range(1, krylov_dim + 1):
        if verbose == 1:
            petscprint(comm, "Arnoldi iteration %d/%d"%(k, krylov_dim))
        v = action(q, v)
        for j in range(k):
            qj = Q.getColumn(j)
            H[j, k - 1] = v.dot(qj)
            v.axpy(-H[j, k - 1], qj)
            Q.restoreColumn(j, qj)
        H[k, k - 1] = v.norm()
        v.scale(1.0 / H[k, k - 1])
        Q.insertVec(k, v)
        v.copy(q)
    q.destroy()
    v.destroy()
    Q.setActiveColumns(0, krylov_dim)
    return (Q, H[:-1,])


def eig(
    L: LinearOperator,
    action: typing.Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec],
    krylov_dim: int,
    n_evals: int,
    process_evals: typing.Optional[
        typing.Callable[[np.ndarray], np.ndarray]
    ] = None,
    verbose: typing.Optional[int] = 0,
) -> typing.Tuple[np.ndarray, SLEPc.BV]:
    r"""
    Compute the eigendecomposition of the linear operator specified by
    :code:`L` and :code:`action`. For example,
    to compute the eigenvalues of :math:`L` closest to the origin,
    set :code:`action = L.solve` and
    :code:`process_evals = lambda x: 1./x`.

    :param L: instance of the :class:`.LinearOperator` class
    :type L: :class:`.LinearOperator`
    :param action: one of :meth:`.LinearOperator.apply`,
        :meth:`.LinearOperator.apply_hermitian_transpose`,
        :meth:`.LinearOperator.solve` or
        :meth:`.LinearOperator.solve_hermitian_transpose`
    :type action: Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec]
    :param krylov_dim: dimension of the Arnoldi Krylov subspace
    :type krylov_dim: int
    :param n_evals: number of eigenvalues to return
    :type n_evals: int
    :param process_evals: function to extract the desired eigenvalues
        (see description above for an example).
    :type process_evals: Optional[Callable[[np.ndarray], np.ndarray]], default
        is :code:`lambda x: x`
    :param verbose: 0 = no printout to terminal, 1 = print progress
    :type verbose: Optional[int], default is 0

    :return: tuple with the desired eigenvalues and corresponding eigenvectors
    :rtype: (numpy.ndarray of size :code:`n_evals x n_evals`,
        SLEPc.BV with :code:`n_evals` columns)
    """
    Q, H = arnoldi_iteration(L, action, krylov_dim, verbose)
    evals, evecs = sp.linalg.eig(H)
    idces = np.flipud(np.argsort(np.abs(evals)))[:n_evals]
    evals = evals[idces]
    evecs = evecs[:, idces]
    evecs_ = PETSc.Mat().createDense(
        evecs.shape, None, evecs, comm=PETSc.COMM_SELF
    )
    Q.multInPlace(evecs_, 0, n_evals)
    Q.setActiveColumns(0, n_evals)
    Q.resize(n_evals, copy=True)
    evals = evals if process_evals is None else process_evals(evals)
    evecs_.destroy()
    return (np.diag(evals), Q)


def match_right_and_left_eigenvectors(
    V: SLEPc.BV, W: SLEPc.BV, Dv: np.ndarray, Dw: np.ndarray
) -> typing.Tuple[SLEPc.BV, SLEPc.BV, np.ndarray, np.ndarray]:
    r"""
    Scale and sort the right and left eigenvectors and corresponding eigenvalues
    of an underlying operator :math:`L`, so that

    .. math::

        W^* L V = D_v = D_w,\quad W^* V = I \in\mathbb{R}^{m\times m}.


    :param V: :code:`m` right eigenvectors
    :type V: SLEPc.BV
    :param W: :code:`m` left eigenvectors
    :type W: SLEPc.BV
    :param Dv: right eigenvalues
    :type Dv: numpy.ndarray of size :code:`m x m`
    :param Dv: eigenvalues computed from the right eigendecomposition
        of :math:`L`
    :type Dv: numpy.ndarray of size :code:`m x m`
    :param Dw: eigenvalues computed from the left eigendecomposition
        of :math:`L`. (Attention: these have already been complex conjugated.)
    :type Dw: numpy.ndarray of size :code:`m x m`

    :return: tuple :math:`(V, W, D_v, D_w)` with the biorthogonalized
        eigenvectors and corresponding eigenvalues
    :rtype: (SLEPc.BV with :code:`m` columns, SLEPc.BV with :code:`m` columns,
        numpy.ndarray of size :code:`m x m`,
        numpy.ndarray of size :code:`m x m`)
    """
    # Match the right and left eigenvalues/vectors
    Dv = np.diag(Dv)
    Dw = np.diag(Dw)
    idces = [np.argmin(np.abs(Dv - val)) for val in Dw]
    Dw = np.diag(Dw[idces])
    Dv = np.diag(Dv)
    Qadj = W.copy()
    for j in range(len(idces)):
        q = Qadj.getColumn(idces[j])
        W.insertVec(j, q)
        Qadj.restoreColumn(idces[j], q)
    Qadj.destroy()
    # Biorthogonalize the eigenvectors
    M = V.dot(W)
    evals, evecs = sp.linalg.eig(M.getDenseArray())
    idces = np.argwhere(np.abs(evals) < 1e-10).reshape(-1)
    evals[idces] += 1e-10
    Minv = evecs @ np.diag(1.0 / evals) @ sp.linalg.inv(evecs)
    Minv = PETSc.Mat().createDense(Minv.shape, None, Minv, PETSc.COMM_SELF)
    V.multInPlace(Minv, 0, V.getSizes()[-1])
    M.destroy()
    Minv.destroy()
    return (V, W, Dv, Dw)


def check_eig_convergence(
    action: typing.Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec],
    D: np.ndarray,
    V: SLEPc.BV,
    monitor: typing.Optional[bool] = False,
) -> np.array:
    r"""
    Check convergence of the eigenpairs by measuring
    :math:`\lVert L v - \lambda v\rVert` for each pair :math:`(\lambda, v)`.

    :param action: one of :meth:`.LinearOperator.apply` or
        :meth:`.LinearOperator.apply_hermitian_transpose`
    :type action: Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec]
    :param D: diagonal 2D numpy array with the eigenvalues
    :type D: numpy.ndarray
    :param V: corresponding eigenvectors
    :type V: SLEPc.BV

    :return: Error vector (each entry is the error of the corresponding
        eigen pair)
    :rtype: np.array
    """
    if monitor:
        petscprint(PETSc.COMM_WORLD, " ")
        petscprint(
            PETSc.COMM_WORLD, "Executing eigenpair convergence check..."
        )
    error_vec = np.zeros(D.shape[0])
    w = V.createVec()
    for j in range(D.shape[-1]):
        v = V.getColumn(j)
        e = v.copy()
        e.scale(D[j, j])
        w = action(v, w)
        e.axpy(-1.0, w)
        error = e.norm()
        error_vec[j] = error.real
        V.restoreColumn(j, v)
        e.destroy()
        if monitor:
            str = "Error for eigenpair %d = %1.15e" % (j + 1, error)
            petscprint(PETSc.COMM_WORLD, str)
    w.destroy()
    if monitor:
        petscprint(
            PETSc.COMM_WORLD, "Executing eigenpair convergence check..."
        )
        petscprint(PETSc.COMM_WORLD, " ")
    return error_vec



__all__ = [
    "bv_add",
    "bv_conj",
    "bv_real",
    "bv_imag",
]

import typing

import numpy as np
from slepc4py import SLEPc
from petsc4py import PETSc

from .matrix import convert_coo_to_csr


def bv_add(alpha: float, X: SLEPc.BV, Y: SLEPc.BV) -> None:
    r"""
    Compute in-place addition :math:`X \leftarrow X + \alpha Y`

    :type alpha: float
    :type X: SLEPc.BV
    :type Y: SLEPc.BV

    :rtype: SLEPc.BV
    """
    Xm = X.getMat()
    Ym = Y.getMat()
    Xm.axpy(alpha, Ym)
    X.restoreMat(Xm)
    Y.restoreMat(Ym)
    return X


def bv_conj(X: SLEPc.BV, inplace: typing.Optional[bool] = False) -> SLEPc.BV:
    r"""
    Returns the complex conjugate :math:`\overline{X}` of the BV structure

    :type X: SLEPc.BV
    :param inplace: in-place if :code:`True`, else the result is stored in a
        new SLEPc.BV structure
    :type inplace: Optional[bool], default is False

    :rtype: SLEPc.BV
    """
    Y = X if inplace else X.copy()
    Ym = Y.getMat()
    Ym.conjugate()
    Y.restoreMat(Ym)
    return Y


def bv_real(X: SLEPc.BV, inplace: typing.Optional[bool] = False) -> SLEPc.BV:
    r"""
    Returns the real part :math:`\text{Re}(X)` of the BV structure

    :type X: SLEPc.BV
    :param inplace: in-place if :code:`True`, else the result is stored in a
        new SLEPc.BV structure
    :type inplace: Optional[bool], default is False

    :rtype: SLEPc.BV
    """
    Y = X if inplace else X.copy()
    Ym = Y.getMat()
    Ym.realPart()
    Y.restoreMat(Ym)
    return Y


def bv_imag(X: SLEPc.BV, inplace: typing.Optional[bool] = False) -> SLEPc.BV:
    r"""
    Returns the imaginary part :math:`\text{Im}(X)` of the BV structure

    :type X: SLEPc.BV
    :param inplace: in-place if :code:`True`, else the result is stored in a
        new SLEPc.BV structure
    :type inplace: Optional[bool], default is False

    :rtype: SLEPc.BV
    """
    Y = X if inplace else X.copy()
    Ym = Y.getMat()
    Ym.imagPart()
    Y.restoreMat(Ym)
    return Y


def bv_slice(
    comm: PETSc.Comm,
    X: SLEPc.BV,
    columns: np.array,
    Y: typing.Optional[SLEPc.BV] = None,
):
    r"""
    Extract a subset of columns from X and store into Y

    :type X: SLEPc.BV
    :param columns: array of columns to extract
    :type columns: np.array
    :type Y: Optional[SLEPc.BV], default is None

    :rtype: SLEPc.BV
    """
    if Y == None:
        Y = SLEPc.BV().create(comm)
        Y.setSizes(X.getSizes()[0], len(columns))
        Y.setType("mat")
    Q = np.zeros((X.getSizes()[-1], len(columns)))
    for i in range(len(columns)):
        Q[columns[i], i] = 1.0
    Q = PETSc.Mat().createDense(Q.shape, None, Q, PETSc.COMM_SELF)
    Y.mult(1.0, 0.0, X, Q)
    Q.destroy()
    return Y


def bv_roll(
    comm: PETSc.Comm,
    X: SLEPc.BV,
    roll: int,
    axis: int,
    in_place: typing.Optional[bool] = False,
):
    r"""
    If :code:`axis=0` roll the rows of X by amount :code:`roll`, if
    :code:`axis=-1` roll the columns of X by amount :code:`roll`. This operation
    can be done in place if :code:`in_place == True` (default is :code:`False`).

    :type X: SLEPc.BV
    :type roll: int
    :type axis: int
    :type in_place: Optional[bool], default is :code:`False`

    :rtype: SLEPc.BV
    """
    Y = X.copy() if not in_place else X

    if axis == -1:
        Q = np.diag(np.ones(Y.getSizes()[-1]))
        Q = np.roll(Q, roll, axis=-1)
        Q = PETSc.Mat().createDense(Q.shape, None, Q, PETSc.COMM_SELF)
        Y.multInPlace(Q, 0, Y.getSizes()[-1])
        Q.destroy()
    else:
        # from ..linear_operators import MatrixLinearOperator
        Ym = Y.getMat()
        r0, r1 = Ym.getOwnershipRange()
        cols = np.arange(r1 - r0) + r0
        rows = np.mod(cols.copy() + roll, Y.getSizes()[0][-1])
        vals = np.ones(len(rows))
        sizes = (Y.getSizes()[0], Y.getSizes()[0])
        row_ptr, col, val = convert_coo_to_csr([rows, cols, vals], sizes)
        M = PETSc.Mat().createAIJ(sizes, comm=comm)
        M.setPreallocationCSR((row_ptr, col))
        M.setValuesCSR(row_ptr, col, val, True)
        M.assemble(False)
        raise ValueError(
            "This error is coming from utils.bv.bv_roll(). "
            + "Need to fix this function because it was causing a"
            + "circular import."
        )
        # Mlop = MatrixLinearOperator(comm, M)
        # Y = Mlop.apply_mat(X, Y)
        # Mlop.destroy()
    return Y



__all__ = [
    "MatrixLinearOperator",
    "LowRankLinearOperator",
    "LowRankUpdatedLinearOperator",
    "ProductLinearOperator",
    "ProjectionLinearOperator",
    "LinearOperator",
]

from .linear_operator import LinearOperator
from .low_rank import LowRankLinearOperator
from .low_rank_updated import LowRankUpdatedLinearOperator
from .matrix import MatrixLinearOperator
from .product import ProductLinearOperator
from .projection import ProjectionLinearOperator

del linear_operator, low_rank, low_rank_updated, matrix, product



__all__ = [
    "create_dense_matrix",
    "create_AIJ_identity",
    "mat_solve_hermitian_transpose",
    "hermitian_transpose",
    "convert_coo_to_csr",
    "assemble_harmonic_resolvent_generator",
]

import numpy as np
from mpi4py import MPI
from petsc4py import PETSc
from typing import Optional

from .miscellaneous import get_mpi_type


def create_dense_matrix(
    comm: PETSc.Comm, sizes: tuple[tuple[int, int], tuple[int, int]]
) -> PETSc.Mat:
    r"""
    Create dense matrix

    :param comm: PETSc communicator
    :type comm: PETSc.Comm
    :param sizes: tuple[tuple[int, int], tuple[int, int]]

    :rtype: PETSc.Mat.Type.DENSE
    """
    M = PETSc.Mat().createDense(sizes, comm=comm)
    M.setUp()
    return M


def create_AIJ_identity(
    comm: PETSc.Comm, sizes: tuple[tuple[int, int], tuple[int, int]]
) -> PETSc.Mat:
    r"""
    Create identity matrix of sparse AIJ type

    :param comm: MPI Communicator
    :type comm: PETSc.Comm
    :param sizes: see `MatSizeSpec <MatSizeSpec_>`_
    :type sizes: tuple[tuple[int, int], tuple[int, int]]

    :return: identity matrix
    :rtype: PETSc.Mat.Type.AIJ
    """
    Id = PETSc.Mat().createConstantDiagonal(sizes, 1.0, comm)
    Id.convert(PETSc.Mat.Type.AIJ)
    return Id


def mat_solve_hermitian_transpose(
    ksp: PETSc.KSP, X: PETSc.Mat, Y: Optional[PETSc.Mat] = None
) -> PETSc.Mat:
    r"""
    Solve :math:`A^{-*}X = Y`, where :math:`X` is a PETSc matrix of type
    :code:`PETSc.Mat.Type.DENSE`

    :param ksp: a KPS solver structure
    :type ksp: PETSc.KSP
    :param X: a dense PETSc matrix
    :type X: PETSc.Mat.Type.DENSE
    :param Y: a dense PETSc matrix
    :type Y: Optional[PETSc.Mat.Type.DENSE] defaults to :code:`None`

    :return: matrix to store the result
    :rtype: PETSc.Mat.Type.DENSE
    """
    sizes = X.getSizes()
    Yarray = np.zeros((sizes[0][0], sizes[-1][-1]), dtype=np.complex128)
    Y = X.duplicate() if Y == None else Y
    y = X.createVecLeft()
    for i in range(X.getSizes()[-1][-1]):
        x = X.getColumnVector(i)
        x.conjugate()
        ksp.solveTranspose(x, y)
        x.conjugate()
        y.conjugate()
        Yarray[:, i] = y.getArray()
        x.destroy()
    y.destroy()
    offset, _ = Y.getOwnershipRange()
    rows = np.arange(Yarray.shape[0], dtype=PETSc.IntType) + offset
    cols = np.arange(Yarray.shape[-1], dtype=PETSc.IntType)
    Y.setValues(rows, cols, Yarray.reshape(-1))
    Y.assemble(None)
    return Y


def hermitian_transpose(
    Mat: PETSc.Mat, in_place=False, MatHT=None
) -> PETSc.Mat:
    r"""
    Return the hermitian transpose of the matrix :code:`Mat`.

    :param Mat: PETSc matrix
    :type Mat: PETSc.Mat
    :param in_place: in-place transposition if :code:`True` and
        out of place otherwise
    :type in_place: Optional[bool] defaults to :code:`False`
    :param MatHT: [optional] matrix with the correct layout to hold the
        hermitian transpose of :code:`Mat`
    :param MatHT: Optional[PETSc.Mat] defaults to :code:`None`
    """
    if in_place == False:
        if MatHT == None:
            sizes = Mat.getSizes()
            MatHT = PETSc.Mat().create(comm=Mat.getComm())
            MatHT.setType(Mat.getType())
            MatHT.setSizes((sizes[-1], sizes[0]))
            MatHT.setUp()
        Mat.setTransposePrecursor(MatHT)
        Mat.hermitianTranspose(MatHT)
        return MatHT
    else:
        MatHT_ = Mat.hermitianTranspose()
        return MatHT_


def convert_coo_to_csr(
    arrays: tuple[np.array, np.array, np.array],
    sizes: tuple[tuple[int, int], tuple[int, int]],
) -> tuple[np.array, np.array, np.array]:
    r"""
    Convert arrays = [row indices, col indices, values] for COO matrix
    assembly to [row pointers, col indices, values] for CSR matrix assembly.
    (Petsc4py currently does not support COO matrix assembly, hence the need
    to convert.)

    :param arrays: a list of numpy arrays (e.g., arrays = [rows,cols,vals])
    :type array: tuple[np.array, np.array, np.array]
    :param sizes: see `MatSizeSpec <MatSizeSpec_>`_
    :type sizes: tuple[np.array, np.array, np.array]

    :return: csr row pointers, column indices and matrix values for CSR
        matrix assembly
    :rtype: tuple[np.array, np.array, np.array]
    """
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    pool = np.arange(comm.Get_size())
    rows, cols, vals = arrays
    idces = np.argsort(rows).reshape(-1)
    rows, cols, vals = rows[idces], cols[idces], vals[idces]

    mat_row_sizes_local = np.asarray(
        comm.allgather(sizes[0][0]), dtype=PETSc.IntType
    )
    mat_row_displ = np.concatenate(([0], np.cumsum(mat_row_sizes_local[:-1])))
    ownership_ranges = np.zeros((comm.Get_size(), 2), dtype=PETSc.IntType)
    ownership_ranges[:, 0] = mat_row_displ
    ownership_ranges[:-1, 1] = ownership_ranges[1:, 0]
    ownership_ranges[-1, 1] = sizes[0][-1]

    send_rows, send_cols = [], []
    send_vals, lengths = [], []
    for i in pool:
        idces = np.argwhere(
            (rows >= ownership_ranges[i, 0]) & (rows < ownership_ranges[i, 1])
        ).reshape(-1)
        lengths.append(np.asarray([len(idces)], dtype=PETSc.IntType))
        send_rows.append(rows[idces])
        send_cols.append(cols[idces])
        send_vals.append(vals[idces])

    recv_bufs = [np.empty(1, dtype=PETSc.IntType) for _ in pool]
    recv_reqs = [comm.Irecv(bf, source=i) for (bf, i) in zip(recv_bufs, pool)]
    send_reqs = [comm.Isend(sz, dest=i) for (i, sz) in enumerate(lengths)]
    MPI.Request.waitall(send_reqs + recv_reqs)
    lengths = [buf[0] for buf in recv_bufs]

    dtypes = [PETSc.IntType, PETSc.IntType, np.complex128]
    my_arrays = []
    for j, array in enumerate([send_rows, send_cols, send_vals]):
        dtype = dtypes[j]
        mpi_type = get_mpi_type(np.dtype(dtype))
        recv_bufs = [
            [np.empty(lengths[i], dtype=dtype), mpi_type] for i in pool
        ]
        recv_reqs = [
            comm.Irecv(bf, source=i) for (bf, i) in zip(recv_bufs, pool)
        ]
        send_reqs = [comm.Isend(array[i], dest=i) for i in pool]
        MPI.Request.waitall(send_reqs + recv_reqs)
        my_arrays.append([recv_bufs[i][0] for i in pool])

    my_rows, my_cols, my_vals = [], [], []
    for i in pool:
        my_rows.extend(my_arrays[0][i])
        my_cols.extend(my_arrays[1][i])
        my_vals.extend(my_arrays[2][i])

    my_rows = (
        np.asarray(my_rows, dtype=PETSc.IntType) - ownership_ranges[rank, 0]
    )
    my_cols = np.asarray(my_cols, dtype=PETSc.IntType)
    my_vals = np.asarray(my_vals, dtype=np.complex128)

    idces = np.argsort(my_rows).reshape(-1)
    my_rows = my_rows[idces]
    my_cols = my_cols[idces]
    my_vals = my_vals[idces]

    ni = 0
    my_rows_ptr = np.zeros(sizes[0][0] + 1, dtype=PETSc.IntType)
    for i in range(sizes[0][0]):
        ni += np.count_nonzero(my_rows == i)
        my_rows_ptr[i + 1] = ni

    return my_rows_ptr, my_cols, my_vals


def assemble_harmonic_resolvent_generator(
    A: PETSc.Mat, freqs: np.array
) -> PETSc.Mat:
    r"""
    Assemble :math:`T = -M + A`, where :math:`A` is the output of
    :func:`resolvent4py.utils.io.read_harmonic_balanced_matrix`
    and :math:`M` is a block
    diagonal matrix with block :math:`k` given by :math:`M_k = i k \omega I`
    and :math:`k\omega` is the :math:`k` th entry of :code:`freqs`.

    :param A: assembled PETSc matrix
    :type A: PETSc.Mat
    :param freqs: array :math:`\omega\left(\ldots, -1, 0, 1, \ldots\right)`
    :type freqs: np.array

    :rtype: PETSc.Mat
    """
    rows_lst = []
    vals_lst = []

    rows = np.arange(*A.getOwnershipRange())
    N = A.getSizes()[0][-1] // len(freqs)
    for i in range(len(freqs)):
        idces = np.intersect1d(rows, np.arange(N * i, N * (i + 1)))
        if len(idces) > 0:
            rows_lst.extend(idces)
            vals_lst.extend(-1j * freqs[i] * np.ones(len(idces)))

    rows = np.asarray(rows_lst, dtype=PETSc.IntType)
    vals = np.asarray(vals_lst, dtype=np.complex128)

    rows_ptr, cols, vals = convert_coo_to_csr([rows, rows, vals], A.getSizes())
    M = PETSc.Mat().create(A.getComm())
    M.setSizes(A.getSizes())
    M.setPreallocationCSR((rows_ptr, cols))
    M.setValuesCSR(rows_ptr, cols, vals, True)
    M.assemble(False)
    M.axpy(1.0, A)
    return M



__all__ = [
    "compute_local_size",
    "sequential_to_distributed_matrix",
    "sequential_to_distributed_vector",
    "distributed_to_sequential_matrix",
    "distributed_to_sequential_vector",
    "scatter_array_from_root_to_all",
]

import typing

import numpy as np
from petsc4py import PETSc
from mpi4py import MPI

from .miscellaneous import get_mpi_type


def compute_local_size(Ng: int) -> int:
    r"""
    Given the global size :code:`Nglob` of a vector, compute
    the local size :code:`Nloc` that will be owned by each processor
    in the MPI pool according to the formula

    .. math::

        N_l = \begin{cases}
            \left\lfloor \frac{N_g}{S} \right\rfloor + 1 & \text{if} 
                \, N_g \text{mod} S > r \\
            \left\lfloor \frac{N_g}{S} \right\rfloor & \text{otherwise}
        \end{cases}

    where :math:`N_g` is the global size, :math:`N_l` is the local size,
    :math:`S` is the size of the MPI pool (i.e., the total number of processors)
    and :math:`r` is the rank of the current processor.

    :param Ng: global size
    :type Ng: int
    
    :return: local size
    :rtype: int
    """
    size, rank = PETSc.COMM_WORLD.getSize(), PETSc.COMM_WORLD.Get_rank()
    Nl = Ng // size + 1 if np.mod(Ng, size) > rank else Ng // size
    return Nl


def sequential_to_distributed_matrix(
    Mat_seq: PETSc.Mat, Mat_dist: PETSc.Mat
) -> PETSc.Mat:
    r"""
    Scatter a sequential dense PETSc matrix (type PETSc.Mat.Type.DENSE) to
    a distributed dense PETSc matrix (type PETSc.Mat.Type.DENSE)

    :param Mat_seq: a sequential dense matrix
    :type Mat_seq: PETSc.Mat.Type.DENSE
    :param Mat_dist: a distributed dense matrix
    :type Mat_dist: PETSc.Mat.Type.DENSE

    :return: a distributed dense matrix
    :rtype: PETSc.Mat.Type.DENSE
    """
    array = Mat_seq.getDenseArray()
    r0, r1 = Mat_dist.getOwnershipRange()
    rows = np.arange(r0, r1)
    cols = np.arange(0, Mat_seq.getSizes()[-1][-1])
    Mat_dist.setValues(rows, cols, array[r0:r1,].reshape(-1))
    Mat_dist.assemble(None)
    return Mat_dist


def sequential_to_distributed_vector(
    vec_seq: PETSc.Vec, vec_dist: PETSc.Vec
) -> PETSc.Vec:
    r"""
    Scatter a sequential PETSc vector (type PETSc.Vec.Type.SEQ) to
    a distributed PETSc vector (type PETSc.Vec.Type.STANDARD)

    :param vec_seq: a sequential vector
    :type vec_seq: PETSc.Vec.Type.SEQ
    :param vec_dist: a distributed vector
    :type vec_dist: PETSc.Vec.Type.STANDARD

    :return: a distributed vector
    :rtype: PETSc.Vec.Type.STANDARD
    """
    array = vec_seq.getArray()
    r0, r1 = vec_dist.getOwnershipRange()
    rows = np.arange(r0, r1)
    vec_dist.setValues(rows, array[r0:r1,])
    vec_dist.assemble(None)
    return vec_dist


def distributed_to_sequential_matrix(Mat_dist: PETSc.Mat) -> PETSc.Mat:
    r"""
    Allgather dense distributed PETSc matrix into a dense sequential PETSc
    matrix.

    :param Mat_dist: a distributed dense matrix
    :type Mat_dist: PETSc.Mat.Type.DENSE

    :rtype: PETSc.Mat.Type.DENSE
    """
    comm = Mat_dist.getComm().tompi4py()
    array = Mat_dist.getDenseArray().copy().reshape(-1)
    counts = np.asarray(comm.allgather(len(array)))
    disps = np.concatenate(([0], np.cumsum(counts[:-1])))
    recvbuf = np.zeros(np.sum(counts), dtype=array.dtype)
    comm.Allgatherv(array, (recvbuf, counts, disps, get_mpi_type(array.dtype)))
    sizes = Mat_dist.getSizes()
    nr, nc = sizes[0][-1], sizes[-1][-1]
    recvbuf = recvbuf.reshape((nr, nc))
    Mat_seq = PETSc.Mat().createDense((nr, nc), None, recvbuf, PETSc.COMM_SELF)
    return Mat_seq


def distributed_to_sequential_vector(vec_dist: PETSc.Vec) -> PETSc.Vec:
    r"""
    Allgather distribued PETSc vector into PETSc sequential vector.

    :param vec_dist: a distributed vector
    :type vec_dist: PETSc.Vec.Type.STANDARD

    :rtype: PETSc.Vec.Type.SEQ
    """
    comm = vec_dist.getComm().tompi4py()
    array = vec_dist.getArray().copy()
    counts = comm.allgather(len(array))
    disps = np.concatenate(([0], np.cumsum(counts[:-1]))).astype(PETSc.IntType)
    recvbuf = np.zeros(np.sum(counts), dtype=array.dtype)
    comm.Allgatherv(array, (recvbuf, counts, disps, get_mpi_type(array.dtype)))
    vec_seq = PETSc.Vec().createWithArray(recvbuf, comm=PETSc.COMM_SELF)
    return vec_seq


def scatter_array_from_root_to_all(
    array: np.array, locsize: typing.Optional[int] = None
) -> np.array:
    r"""
    Scatter numpy array from root to all other processors

    :param array: numpy array to be scattered
    :type array: numpy.array
    :param locsize: local size owned by each processor. If :code:`None`,
        this is computed using the same logic as in the
        :func:`.compute_local_size` routine.
    :type locsize: Optional[int] default to None

    :rtype: numpy.array
    """
    comm = MPI.COMM_WORLD
    size, rank = comm.Get_size(), comm.Get_rank()
    counts, displs = None, None
    if locsize == None:
        if rank == 0:
            n = len(array)
            counts = np.asarray(
                [
                    n // size + 1 if np.mod(n, size) > j else n // size
                    for j in range(size)
                ]
            )
            displs = np.concatenate(([0], np.cumsum(counts[:-1])))
            count = counts[0]
            dtype = array.dtype
            for j in range(1, size):
                comm.send(counts[j], dest=j, tag=0)
                comm.send(dtype, dest=j, tag=1)
        else:
            count = comm.recv(source=0, tag=0)
            dtype = comm.recv(source=0, tag=1)
    else:
        count = locsize
        counts = comm.gather(locsize, root=0)
        if rank == 0:
            dtype = array.dtype
            displs = np.concatenate(([0], np.cumsum(counts[:-1])))
            for j in range(1, size):
                comm.send(dtype, dest=j, tag=1)
        else:
            dtype = comm.recv(source=0, tag=1)
    recvbuf = np.empty(count, dtype=dtype)
    counts = (
        np.asarray(counts, dtype=PETSc.IntType)
        if counts is not None
        else counts
    )
    displs = (
        np.asarray(displs, dtype=PETSc.IntType)
        if displs is not None
        else displs
    )
    comm.Scatterv(
        [array, counts, displs, get_mpi_type(dtype)], recvbuf, root=0
    )
    return recvbuf



__all__ = [
    "randomized_svd",
    "check_randomized_svd_convergence",
]

import typing

import numpy as np
import scipy as sp
from petsc4py import PETSc
from slepc4py import SLEPc

from ..linear_operators import LinearOperator
from ..utils.matrix import create_dense_matrix
from ..utils.miscellaneous import petscprint
from ..utils.vector import enforce_complex_conjugacy


def randomized_svd(
    L: LinearOperator,
    action: typing.Callable[[SLEPc.BV, SLEPc.BV], SLEPc.BV],
    n_rand: int,
    n_loops: int,
    n_svals: int,
    verbose: typing.Optional[int] = 0,
) -> typing.Tuple[SLEPc.BV, np.ndarray, SLEPc.BV]:
    r"""
    Compute the singular value decomposition (SVD) of the linear operator
    specified by :code:`L` and :code:`action` using a randomized SVD algorithm.
    (See [Halko2011]_.)
    For example, with :code:`L.solve_mat` we compute

    .. math::

        L^{-1} = U \Sigma V^*.


    :param L: instance of the :class:`.LinearOperator` class
    :type L: :class:`.LinearOperator`
    :param action: one of :meth:`.LinearOperator.apply_mat` or
        :meth:`.LinearOperator.solve_mat`
    :type action: Callable[[SLEPc.BV, SLEPc.BV], SLEPc.BV]
    :param n_rand: number of random vectors
    :type n_rand: int
    :param n_loops: number of randomized svd power iterations
        (see [Ribeiro2020]_ for additional details on this parameter)
    :type n_loops: int
    :param n_svals: number of singular triplets to return
    :type n_svals: int
    :param verbose: defines verbosity of output to terminal (useful to
        monitor progress during time stepping). = 0 no printout to terminal,
        = 1 monitor randomized SVD iterations.
    :type verbose: Optional[int], default is 0

    :return: leading :code:`n_svals` left singular vectors,
        singular values and right singular vectors
    :rtype: Tuple[SLEPc.BV, np.ndarray, SLEPc.BV]

    References
    ----------
    .. [Halko2011] Halko et al., *Finding Structure With Randomness:
        Probabilistic Algorithms For Constructing Approximate Matrix
        Decompositions*, SIAM Review, 2011
    .. [Ribeiro2020] Ribeiro et al., *Randomized resolvent analysis*,
        Physical Review Fluids, 2020
    """
    comm = L.get_comm()
    if action != L.apply_mat and action != L.solve_mat:
        raise ValueError(f"action must be L.apply_mat or L.solve_mat.")
    action_adj = (
        L.apply_hermitian_transpose_mat
        if action == L.apply_mat
        else L.solve_hermitian_transpose_mat
    )
    # Assemble random BV (this will be multiplied against L^*)
    rowsizes = L.get_dimensions()[0]
    # Set seed
    rank = L.get_comm().getRank()
    rand = PETSc.Random().create(comm=comm)
    rand.setType(PETSc.Random.Type.RAND)
    rand.setSeed(round(np.random.randint(1000, 100000) + rank))
    X = SLEPc.BV().create(comm=comm)
    X.setSizes(rowsizes, n_rand)
    X.setType("mat")
    X.setRandomContext(rand)
    X.setRandomNormal()
    rand.destroy()
    for j in range(n_rand):
        xj = X.getColumn(j)
        if L.get_real_flag():
            row_offset = xj.getOwnershipRange()[0]
            rows = np.arange(rowsizes[0], dtype=PETSc.IntType) + row_offset
            xj.setValues(rows, xj.getArray().real)
            xj.assemble()
        if L.get_block_cc_flag():
            enforce_complex_conjugacy(comm, xj, L.get_nblocks())
        X.restoreColumn(j, xj)
    X.orthogonalize(None)
    # Perform randomized SVD loop
    Qadj = SLEPc.BV().create(comm=comm)
    Qadj.setSizes(L.get_dimensions()[-1], n_rand)
    Qadj.setType("mat")
    Qadj = action_adj(X, Qadj)
    Qadj.orthogonalize(None)
    X.destroy()
    Qfwd = SLEPc.BV().create(comm=comm)
    Qfwd.setSizes(L.get_dimensions()[0], n_rand)
    Qfwd.setType("mat")
    R = create_dense_matrix(PETSc.COMM_SELF, (n_rand, n_rand))
    for j in range(n_loops):
        if verbose == 1:
            str = "Loop %d/%d, forward action"%(j+1, n_loops)
            petscprint(comm, str)
        Qfwd = action(Qadj, Qfwd)
        Qfwd.orthogonalize(None)
        if verbose == 1:
            str = "Loop %d/%d, adjoint action"%(j+1, n_loops)
            petscprint(comm, str)
        Qadj = action_adj(Qfwd, Qadj)
        Qadj.orthogonalize(R)
    # Compute low-rank SVD
    u, s, v = sp.linalg.svd(R.getDenseArray())
    R.destroy()
    v = v.conj().T
    s = s[:n_svals]
    u = u[:, :n_svals]
    v = v[:, :n_svals]
    u = PETSc.Mat().createDense(
        (n_rand, n_svals), None, u, comm=PETSc.COMM_SELF
    )
    v = PETSc.Mat().createDense(
        (n_rand, n_svals), None, v, comm=PETSc.COMM_SELF
    )
    Qfwd.multInPlace(v, 0, n_svals)
    Qfwd.setActiveColumns(0, n_svals)
    Qfwd.resize(n_svals, copy=True)
    Qadj.multInPlace(u, 0, n_svals)
    Qadj.setActiveColumns(0, n_svals)
    Qadj.resize(n_svals, copy=True)
    u.destroy()
    v.destroy()
    return (Qfwd, np.diag(s), Qadj)


def check_randomized_svd_convergence(
    action: typing.Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec],
    U: SLEPc.BV,
    S: np.ndarray,
    V: SLEPc.BV,
    monitor: typing.Optional[bool] = False,
) -> np.array:
    r"""
    Check the convergence of the singular value triplets by measuring
    :math:`\lVert Av/\sigma - u\rVert` for every triplet :math:`(u, \sigma, v)`.

    :param action: one of :meth:`.LinearOperator.apply` or
        :meth:`.LinearOperator.solve`
    :type action: Callable[[PETSc.Vec, PETSc.Vec], PETSc.Vec]
    :param U: left singular vectors
    :type U: SLEPc.BV
    :param D: diagonal 2D numpy array with the singular values
    :type D: numpy.ndarray
    :param V: right singular vectors
    :type V: SLEPc.BV

    :return: Error vector (each entry is the error of the corresponding
        singular triplet)
    :rtype: np.array
    """
    if monitor:
        petscprint(PETSc.COMM_WORLD, " ")
        petscprint(
            PETSc.COMM_WORLD, "Executing SVD triplet convergence check..."
        )
    x = U.createVec()
    n_svals = S.shape[-1]
    error_vec = np.zeros(n_svals)
    for k in range(n_svals):
        v = V.getColumn(k)
        u = U.getColumn(k)
        x = action(v, x)
        x.scale(1.0 / S[k, k])
        x.axpy(-1.0, u)
        error = x.norm()
        error_vec[k] = error.real
        if monitor:
            str = "Error for SVD triplet %d = %1.15e" % (k + 1, error)
            petscprint(PETSc.COMM_WORLD, str)
        U.restoreColumn(k, u)
        V.restoreColumn(k, v)
    x.destroy()
    if monitor:
        petscprint(
            PETSc.COMM_WORLD, "Executing SVD triplet convergence check..."
        )
        petscprint(PETSc.COMM_WORLD, " ")



__all__ = ["resolvent_analysis_rsvd_dt"]

import typing

import math
import numpy as np
import scipy as sp
import time as tlib
from petsc4py import PETSc
from slepc4py import SLEPc

from ..linear_operators import LinearOperator
from ..utils.matrix import create_dense_matrix
from ..utils.miscellaneous import petscprint
from ..utils.vector import vec_real


def _reorder_list(Qlist: list[SLEPc.BV], Qlist_reordered: list[SLEPc.BV]):
    for j in range(Qlist[0].getSizes()[-1]):
        for i in range(len(Qlist)):
            Qij = Qlist[i].getColumn(j)
            Qlist_reordered[j].insertVec(i, Qij)
            Qlist[i].restoreColumn(j, Qij)
    return Qlist_reordered


def _ifft(
    Xhat: SLEPc.BV,
    x: PETSc.Vec,
    omegas: np.array,
    t: float,
    adjoint: typing.Optional[bool] = False,
):
    sign = -1 if adjoint else 1
    q = np.exp(1j * omegas * t * sign)
    if np.min(omegas) == 0.0:
        c = 2 * np.ones(len(q))
        c[0] = 1.0
        q *= c
        Xhat.multVec(1.0, 0.0, x, q)
        x = vec_real(x, True)
    else:
        Xhat.multVec(1.0, 0.0, x, q)
    return x


def _fft(
    X: SLEPc.BV,
    Xhat: SLEPc.BV,
    real: typing.Optional[bool] = True,
    adjoint: typing.Optional[bool] = False,
):
    n_omegas = Xhat.getSizes()[-1]
    n_tstore = X.getSizes()[-1]

    Xhat_mat = Xhat.getMat()
    Xhat_mat_a = Xhat_mat.getDenseArray()
    Xmat = X.getMat()
    Xmat_a = Xmat.getDenseArray().copy()
    Xmat_a = Xmat_a.conj() if adjoint else Xmat_a
    if real:
        Xhat_mat_a[:, :] = (
            np.fft.rfft(Xmat_a.real, axis=-1)[:, :n_omegas] / n_tstore
        )
    else:
        n_omegas = int((n_omegas - 1) // 2 + 1)
        idces_pos = np.arange(n_omegas)
        idces_neg = np.arange(-n_omegas + 1, 0)
        idces = np.concatenate((idces_pos, idces_neg))
        Xhat_mat_a[:, :] = np.fft.fft(Xmat_a, axis=-1)[:, idces] / n_tstore

    Xhat_mat_a[:, :] = Xhat_mat_a.conj() if adjoint else Xhat_mat_a
    X.restoreMat(Xmat)
    Xhat.restoreMat(Xhat_mat)
    return Xhat


def _action(
    L: LinearOperator,
    Laction: typing.Callable,
    tsim: np.array,
    tstore: np.array,
    omegas: np.array,
    x: PETSc.Vec,
    Fhat: SLEPc.BV,
    Xhat: SLEPc.BV,
    X: SLEPc.BV,
    tol: typing.Optional[float] = 1e-3,
    verbose: typing.Optional[int] = 0,
):
    dt = tsim[1] - tsim[0]
    dt_store = tstore[1] - tstore[0]
    T = tstore[-1] + dt_store
    n_save = round(dt_store / dt)
    n_save_per_period = round(T / dt_store)
    n_periods = round((tsim[-1] + dt) / T)

    rhs = L.create_left_vector()
    rhs_im1 = rhs.copy()
    rhs_temp = rhs.copy()
    Lx = x.copy()
    x0 = x.copy()

    adjoint = True if Laction == L.apply_hermitian_transpose else False
    save_idx = 0
    period = 0
    X.insertVec(0, x)
    for i in range(1, len(tsim)):
        rhs = _ifft(Fhat, rhs, omegas, tsim[i - 1], adjoint)
        rhs.axpy(1.0, Laction(x, Lx))
        if i == 1:
            rhs.copy(rhs_im1)
        else:
            rhs.copy(rhs_temp)
            rhs.scale(3 / 2)
            rhs.axpy(-1 / 2, rhs_im1)
            rhs_temp.copy(rhs_im1)
        x.axpy(dt, rhs)

        if np.mod(i, n_save) == 0:
            if math.isnan(x.norm()):
                raise ValueError(f"Code blew up at time step {i}")
            save_idx = np.mod(save_idx + 1, n_save_per_period)
            if save_idx == 0:
                period += 1
                x0.axpy(-1.0, x)
                error = 100 * x0.norm() / x.norm()
                x.copy(x0)
                if verbose > 1:
                    str = (
                        "Deviation from periodicity at period %d/%d = %1.5e"
                        % (
                            period,
                            n_periods,
                            error,
                        )
                    )
                    petscprint(PETSc.COMM_WORLD, str)
                if (
                    error < tol
                ):  # Reached limit cycle, no need to integrate further
                    break
            X.insertVec(save_idx, x)

    Xhat = _fft(X, Xhat, L.get_real_flag(), adjoint)
    objs = [rhs, rhs_im1, rhs_temp, Lx, x0]
    for obj in objs:
        obj.destroy()

    return Xhat


def _create_time_and_frequency_arrays(
    dt: float, omega: float, n_omegas: int, n_periods: int, real: bool
):
    T = 2 * np.pi / omega
    tstore = np.linspace(0, T, num=2 * n_omegas + 4, endpoint=False)
    dt_store = tstore[1] - tstore[0]
    dt = dt_store / round(dt_store / dt)
    n_tsteps_per_period = round(T / dt)
    tsim = dt * np.arange(0, n_periods * n_tsteps_per_period)
    nsave = round(dt_store / dt)
    tsim_check = tsim[(n_periods - 1) * n_tsteps_per_period :: nsave].copy()
    tsim_check -= tsim[(n_periods - 1) * n_tsteps_per_period]
    if np.linalg.norm(tsim_check - tstore) >= 1e-10:
        raise ValueError("Simulation and storage times are not matching.")

    omegas = np.arange(n_omegas + 1) * omega
    omegas = (
        omegas if real else np.concatenate((omegas, -np.flipud(omegas[1:])))
    )

    return tsim, tstore, omegas, len(omegas)


def resolvent_analysis_rsvd_dt(
    L: LinearOperator,
    dt: float,
    omega: float,
    n_omegas: int,
    n_periods: int,
    n_rand: int,
    n_loops: int,
    n_svals: int,
    tol: typing.Optional[float] = 1e-3,
    verbose: typing.Optional[int] = 0,
) -> typing.Tuple[SLEPc.BV, np.ndarray, SLEPc.BV]:
    r"""
    Perform resolvent analysis using randomized linear algebra and time
    stepping.
    In particular, it can be shown that

    .. math::

        x_\omega e^{i\omega t} = \left(i\omega I - A\right)^{-1}f_\omega
            e^{i\omega t}
            \to \int_0^t e^{A(t-\tau)}f_\omega e^{i\omega\tau} d\tau

    for sufficiently long time :math:`t \gg 1` and for any complex-valued
    forcing :math:`f(t) = f_\omega e^{i\omega t}` (assuming that :math:`A` is
    stable).
    Computing the integral on the right-hand side can be done by integrating

    .. math::

        \frac{d}{dt}x(t) = Ax(t) + f(t)

    forward in time with initial condition :math:`x(0) = 0`.
    Thus, the action of the resolvent operator
    :math:`R(i\omega) = \left(i\omega I - A\right)^{-1}` on a vector
    :math:`f_\omega` can be computed by time-stepping the ODE above.
    For now, time integration is performed explicitly via the
    Adams-Bashforth scheme.
    (See [Martini2021]_ and [Farghadan2025]_ for more details.)

    .. note::

        This function is meant for systems whose dimension is so large
        that linear systems of the form
        :math:`(i\omega I - A)x_\omega = f_\omega`
        cannot be solved easily. Typically, this happens when
        :math:`\text{dim}(x_\omega) \sim O(10^7)` or larger.
        If you have a "small enough" system, then we highly recommend using
        :func:`.randomized_svd.randomized_svd` instead for the singular
        value decomposition of the resolvent.

    :param L: linear operator representing :math:`A`
    :type L: LinearOperator

    :param dt: target time step :math:`\Delta t` to integrate the ODE
    :type dt: float

    :param omega: fundamental frequency :math:`\omega` of the forcing
    :type omega: float

    :param n_omegas: number of integer harmonics of :math:`\omega` to resolve
    :type n_omegas: int

    :param n_periods: number of periods :math:`T = 2\pi/\omega` to integrate
        the ODE through. (This number should be large enough that we can
        expect transients to have decayed.)
    :type n_periods: int

    :param n_rand: number of random forcing vectors for randomized SVD
    :type n_rand: int

    :param n_loops: number of power iterations for randomized SVD
    :type n_loops: int

    :param n_svals: number of singular values/vectors to output
    :type n_svals: int

    :param tol: integrate the ODE forward for :code:`n_periods` or until
        :math:`\lVert x(kT) - x((k-1)T) \rVert < \mathrm{tol}`.
    :type tol: Optional[float], default is :math:`10^{-3}`

    :param verbose: defines verbosity of output to terminal (useful to
        monitor progress during time stepping). = 0 no printout to terminal,
        = 1 monitor randomized SVD iterations, = 2 monitor randomized SVD
        iterations and time-stepping progress.
    :type verbose: Optional[int], default is 0

    :return: left singular vectors, singular values, and right singular vectors
        of the resolvent operators :math:`R(i\omega)` evaluated at frequencies
        :math:`\Omega = \omega\{0, 1, 2, \ldots, n_{\omega}\}` if
        the linear operator :math:`A` is real-valued; otherwise at frequencies
        :math:`\Omega = \omega\{0, 1, 2, \ldots, n_{\omega}, -n_{\omega},-(n_{\omega}-1) \ldots, -1\}`
    :rtype: Tuple[List[SLEPc.BV], List[np.ndarray], List[SLEPc.BV]]

    References
    ----------
    .. [Martini2021] Martini et al., *Efficient computation of global
        resolvent modes*, Journal of Fluid Mechanics, 2021
    .. [Farghadan2025] Farghadan et al., *Scalable resolvent analysis
        for three-dimensional flows*, Journal of Computational Physics, 2025
    """

    size = L.get_dimensions()[0]

    tsim, tstore, omegas, n_omegas = _create_time_and_frequency_arrays(
        dt, omega, n_omegas, n_periods, L.get_real_flag()
    )

    Qadj_hat_lst, Qfwd_hat_lst = [], []
    for _ in range(n_rand):
        # Set seed
        rank = L.get_comm().getRank()
        rand = PETSc.Random().create(comm=L.get_comm())
        rand.setType(PETSc.Random.Type.RAND)
        rand.setSeed(round(np.random.randint(1000, 100000) + rank))
        # Initialize Qadj_hat and Qfwd_hat with random BVs of size N x n_rand
        X = SLEPc.BV().create(comm=L.get_comm())
        X.setSizes(size, n_omegas)
        X.setType("mat")
        X.setRandomContext(rand)
        X.setRandomNormal()
        rand.destroy()
        if L.get_real_flag():
            v = X.getColumn(0)
            v = vec_real(v, True)
            X.restoreColumn(0, v)
        Qadj_hat_lst.append(X.copy())
        Qfwd_hat_lst.append(X.copy())
        X.destroy()

    # Initialize Qadj_hat and Qfwd_hat with BVs of size N x n_omegas
    X = SLEPc.BV().create(comm=L.get_comm())
    X.setSizes(size, n_rand)
    X.setType("mat")
    Qadj_hat_lst2 = [X.copy() for _ in range(n_omegas)]
    Qfwd_hat_lst2 = [X.copy() for _ in range(n_omegas)]
    X.destroy()

    Qadj = SLEPc.BV().create(comm=L.get_comm())
    Qadj.setSizes(size, len(tstore))
    Qadj.setType("mat")
    Qfwd = Qadj.duplicate()

    Qadj_hat_lst2 = _reorder_list(Qadj_hat_lst, Qadj_hat_lst2)
    for i in range(len(Qadj_hat_lst2)):
        Qadj_hat_lst2[i].orthogonalize(None)
    Qadj_hat_lst = _reorder_list(Qadj_hat_lst2, Qadj_hat_lst)

    x = L.create_left_vector()
    for j in range(n_loops):
        for k in range(n_rand):
            if verbose > 0:
                str = "Loop %d/%d, random vector %d/%d (forward action)" % (
                    j + 1,
                    n_loops,
                    k + 1,
                    n_rand,
                )
                petscprint(L.get_comm(), str)
            x.zeroEntries()
            Qfwd_hat_lst[k] = _action(
                L,
                L.apply,
                tsim,
                tstore,
                omegas,
                x,
                Qadj_hat_lst[k],
                Qfwd_hat_lst[k],
                Qfwd,
                tol,
                verbose,
            )
        Qfwd_hat_lst2 = _reorder_list(Qfwd_hat_lst, Qfwd_hat_lst2)
        for i in range(len(Qfwd_hat_lst2)):
            Qfwd_hat_lst2[i].orthogonalize(None)
        Qfwd_hat_lst = _reorder_list(Qfwd_hat_lst2, Qfwd_hat_lst)

        for k in range(n_rand):
            if verbose > 0:
                str = "Loop %d/%d, random vector %d/%d (adjoint action)" % (
                    j + 1,
                    n_loops,
                    k + 1,
                    n_rand,
                )
                petscprint(L.get_comm(), str)
            x.zeroEntries()
            Qadj_hat_lst[k] = _action(
                L,
                L.apply_hermitian_transpose,
                tsim,
                tstore,
                omegas,
                x,
                Qfwd_hat_lst[k],
                Qadj_hat_lst[k],
                Qadj,
                tol,
                verbose,
            )
        Qadj_hat_lst2 = _reorder_list(Qadj_hat_lst, Qadj_hat_lst2)
        Rlst = []
        R = create_dense_matrix(PETSc.COMM_SELF, (n_rand, n_rand))
        for i in range(len(Qadj_hat_lst2)):
            Qadj_hat_lst2[i].orthogonalize(R)
            Rlst.append(R.copy())
        R.destroy()
        Qadj_hat_lst = _reorder_list(Qadj_hat_lst2, Qadj_hat_lst)
        if j < n_loops - 1:
            for obj in Rlst:
                obj.destroy()

    x.destroy()
    # Compute low-rank SVD
    Slst = []
    for j, R in enumerate(Rlst):
        u, s, v = sp.linalg.svd(R.getDenseArray())
        v = v.conj().T
        s = s[:n_svals]
        u = u[:, :n_svals]
        v = v[:, :n_svals]
        u = PETSc.Mat().createDense(
            (n_rand, n_svals), None, u, comm=PETSc.COMM_SELF
        )
        v = PETSc.Mat().createDense(
            (n_rand, n_svals), None, v, comm=PETSc.COMM_SELF
        )
        Qfwd_hat_lst2[j].multInPlace(v, 0, n_svals)
        Qfwd_hat_lst2[j].setActiveColumns(0, n_svals)
        Qfwd_hat_lst2[j].resize(n_svals, copy=True)
        Qadj_hat_lst2[j].multInPlace(u, 0, n_svals)
        Qadj_hat_lst2[j].setActiveColumns(0, n_svals)
        Qadj_hat_lst2[j].resize(n_svals, copy=True)
        Slst.append(np.diag(s))
        u.destroy()
        v.destroy()

    lists = [Rlst, Qfwd_hat_lst, Qadj_hat_lst]
    for lst in lists:
        for obj in lst:
            obj.destroy()

    return Qfwd_hat_lst2, Slst, Qadj_hat_lst2



__all__ = [
    "generate_random_petsc_sparse_matrix",
    "generate_random_petsc_vector",
]

import typing

import numpy as np
import scipy as sp
from mpi4py import MPI
from petsc4py import PETSc

from .comms import scatter_array_from_root_to_all
from .matrix import convert_coo_to_csr


def generate_random_petsc_sparse_matrix(
    sizes: typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]],
    nnz: int,
    complex: typing.Optional[bool] = False,
) -> PETSc.Mat:
    r"""
    :param sizes: :code:`((local rows, global rows), (local cols, global cols))`
    :type sizes: Tuple[Tuple[int, int], Tuple[int, int]]
    :param nnz: number of non-zero entries
    :type nnz: int
    :param complex: :code:`True` if you want a complex-valued matrix
    :type complex: Optional[bool], default is False

    :return: a sparse PETSc matrix
    :rtype: PETSc.Mat
    """
    comm = PETSc.COMM_WORLD
    rank = comm.getRank()
    nrows, ncols = sizes[0][-1], sizes[-1][-1]
    # Generate random matrix on root
    arrays = [None, None, None]
    if rank == 0:
        dtype = np.complex128 if complex else np.float64
        A = sp.sparse.random(
            nrows,
            ncols,
            density=nnz / (nrows * ncols),
            format="csr",
            dtype=dtype,
        )
        if nrows == ncols:  # add identity to make A invertible
            A += sp.sparse.identity(nrows, dtype=dtype, format="csr")
        A = A.tocoo()
        arrays = [A.row, A.col, A.data + 1j * 0]  # need complex type for PETSc
    # Scatter to all other processors and assemble
    recv_bufs = [scatter_array_from_root_to_all(a) for a in arrays]
    row_ptrs, cols, data = convert_coo_to_csr(recv_bufs, sizes)
    A = PETSc.Mat().create(comm=comm)
    A.setSizes(sizes)
    A.setUp()
    A.setPreallocationCSR((row_ptrs, cols, data))
    A.setValuesCSR(row_ptrs, cols, data)
    A.assemble(None)
    return A


def generate_random_petsc_vector(
    sizes: typing.Tuple[int, int],
    complex: typing.Optional[bool] = False,
) -> PETSc.Vec:
    r"""
    :param sizes: vector sizes
    :type sizes: Tuple[int, int]
    :param complex: :code:`True` if you want a complex-valued vector
    :type complex: Optional[bool], default is False

    :rtype: PETSc.Vec
    """
    comm = PETSc.COMM_WORLD
    array = None
    if comm.getRank() == 0:
        vec = np.random.randn(sizes[-1])
        array = vec + 1j * np.random.randn(sizes[-1]) if complex else vec
    array = scatter_array_from_root_to_all(array, sizes[0])
    vec = PETSc.Vec().createWithArray(array, sizes, None, comm=comm)
    return vec



import abc
from typing import Optional, Union

import numpy as np
from petsc4py import PETSc
from slepc4py import SLEPc

from ..utils.errors import raise_not_implemented_error
from ..utils.random import generate_random_petsc_vector
from ..utils.vector import check_complex_conjugacy, enforce_complex_conjugacy


class LinearOperator(metaclass=abc.ABCMeta):
    r"""
    Abstract base class for linear operators :math:`L`.

    :param comm: MPI communicator :code:`PETSc.COMM_WORLD`
    :type comm: PETSc.Comm
    :param name: name of the linear operator
    :type name: str
    :param dimensions: local and global sizes of the range and domain
        of the linear operator, :code:`[[local rows, global rows], 
        [local columns, global columns]]`
    :type dimensions: tuple[tuple[int, int], tuple[int, int]]
    :param nblocks: number of blocks (if the linear operator has block \
        structure)
    :type nblocks: Optional[Unions[int, None]], default is None
    """

    def __init__(
        self: "LinearOperator",
        comm: PETSc.Comm,
        name: str,
        dimensions: tuple[tuple[int, int], tuple[int, int]],
        nblocks: Optional[Union[int, None]] = None,
    ) -> None:
        self._comm = comm
        self._name = name
        self._dimensions = dimensions
        self._nblocks = nblocks
        self._real_flag = self.check_if_real_valued()
        self._block_cc_flag = (
            self.check_if_complex_conjugate_structure()
            if self.get_nblocks() != None
            else None
        )

    def get_comm(self: "LinearOperator") -> PETSc.Comm:
        r"""
        The MPI communicator

        :rtype: PETSc.Comm
        """
        return self._comm

    def get_name(self: "LinearOperator") -> str:
        r"""
        The name of the linear operator

        :rtype: str
        """
        return self._name

    def get_dimensions(
        self: "LinearOperator",
    ) -> tuple[tuple[int, int], tuple[int, int]]:
        r"""
        The local and global dimensions of the linear operator

        :rtype: tuple[tuple[int, int], tuple[int, int]]
        """
        return self._dimensions

    def get_nblocks(self: "LinearOperator") -> Union[int, None]:
        r"""
        The number of blocks of the linear operator

        :rtype: Union[int, None]
        """
        return self._nblocks

    def get_real_flag(self: "LinearOperator") -> bool:
        r"""
        :code:`True` if the operator is real-valued, :code:`False`
        otherwise.

        :rtype: bool
        """
        return self._real_flag

    def get_block_cc_flag(self: "LinearOperator") -> Union[bool, None]:
        r"""
        Returns :code:`True` if the operator has complex-conjugate block
        structure (like in the harmonic-balancing operator). :code:`False` if
        the operator does not have complex-conjugate block structure,
        :code:`None` if :code:`.get_nblocks` returns :code:`None`.

        :rtype: Union[bool, None]
        """
        return self._block_cc_flag

    def create_right_vector(self: "LinearOperator") -> PETSc.Vec:
        r"""
        :return: a PETSc vector that :math:`L` can be multiplied against
        :rtype: PETSc.Vec
        """
        vec = PETSc.Vec().create(comm=self.get_comm())
        vec.setSizes(self.get_dimensions()[-1])
        vec.setType("standard")
        return vec

    def create_right_bv(self: "LinearOperator", ncols: int) -> SLEPc.BV:
        r"""
        :param ncols: number of columns in the BV
        :type ncols: int

        :return: a SLEPc BV that :math:`L` can be multiplied against
        :rtype: SLEPc.BV
        """
        bv = SLEPc.BV().create(comm=self.get_comm())
        bv.setSizes(self.get_dimensions()[-1], ncols)
        bv.setType("mat")
        return bv

    def create_left_vector(self: "LinearOperator") -> PETSc.Vec:
        r"""
        :return: a PETSc vector where :math:`Lx` can be stored into
        :rtype: PETSc.Vec
        """
        vec = PETSc.Vec().create(comm=self.get_comm())
        vec.setSizes(self.get_dimensions()[0])
        vec.setType("standard")
        return vec

    def create_left_bv(self: "LinearOperator", ncols: int) -> SLEPc.BV:
        r"""
        :param ncols: number of columns in the BV
        :param type: int

        :return: a SLEPc BV where :math:`LX` can be stored into
        :rtype: SLEPc.BV
        """
        bv = SLEPc.BV().create(comm=self.get_comm())
        bv.setSizes(self.get_dimensions()[0], ncols)
        bv.setType("mat")
        return bv

    def check_if_real_valued(self: "LinearOperator") -> bool:
        r"""
        :return: :code:`True` if the linear operator is real-valued,
            :code:`False` otherwise
        :rtype: bool
        """
        sizes = self.get_dimensions()[-1]
        x = generate_random_petsc_vector(sizes)
        Lx = self.apply(x)
        Lxai = Lx.getArray().imag
        norm = np.sqrt(
            sum(
                self.get_comm().tompi4py().allgather(np.linalg.norm(Lxai) ** 2)
            )
        )
        result = True if norm <= 1e-14 else False
        x.destroy()
        Lx.destroy()
        return result

    def check_if_complex_conjugate_structure(self: "LinearOperator") -> bool:
        r"""
        Given a vector

        .. math::
            x = \left(\ldots,x_{-1},x_{0},x_{1},\ldots\right)

        with vector-valued entries that satisfy \
        :math:`x_{-i} = \overline{x_i}`, check if the vector :math:`Lx` \
        satisfies :math:`(Lx)_{-i}=\overline{(Lx)_{i}}`. (Here, the \
        overline denote complex conjugation.)

        :return: :code:`True` if the linear operator has complex-conjugate
            structure, :code:`False` otherwise.
        :rtype: bool
        """
        x = generate_random_petsc_vector(self.get_dimensions()[-1])
        enforce_complex_conjugacy(self.get_comm(), x, self.get_nblocks())
        cc_x = check_complex_conjugacy(self.get_comm(), x, self.get_nblocks())
        if cc_x == False:
            raise ValueError(
                f"Error from {self.get_name()}.check_if_complex_conjugate"
                f"_structure(): complex conjugacy was not enforced "
                f"appropriately."
            )
        Lx = self.apply(x)
        result = check_complex_conjugacy(
            self.get_comm(), Lx, self.get_nblocks()
        )
        x.destroy()
        Lx.destroy()
        return result

    # Methods that must be implemented by subclasses
    @abc.abstractmethod
    def apply(
        self: "LinearOperator", x: PETSc.Vec, y: Optional[PETSc.Vec] = None
    ) -> PETSc.Vec:
        r"""
        Compute :math:`y = Lx`.

        :param x: a PETSc vector
        :type x: PETSc.Vec
        :param y: a PETSc vector to store the result
        :type y: Optional[PETSc.Vec], default is None

        :rtype: PETSc.Vec
        """

    @abc.abstractmethod
    def apply_mat(
        self: "LinearOperator", X: SLEPc.BV, Y: Optional[SLEPc.BV] = None
    ) -> SLEPc.BV:
        r"""
        Compute :math:`Y = LX`

        :param X: a SLEPc BV
        :type X: SLEPc.BV
        :param Y: a SLEPc BV to store the result
        :type Y: Optional[SLEPc.BV], default is None

        :rtype: SLEPc.BV
        """

    @abc.abstractmethod
    def destroy(self: "LinearOperator") -> None:
        r"""
        Destroy the PETSc and SLEPc objects associated with :math:`L`
        """

    # Methods that don't necessarily need to be implemented by subclasses
    @raise_not_implemented_error
    def apply_hermitian_transpose(
        self: "LinearOperator", x: PETSc.Vec, y: Optional[PETSc.Vec] = None
    ) -> PETSc.Vec:
        r"""
        Compute :math:`y = L^*x`.

        :param x: a PETSc vector
        :type x: PETSc.Vec
        :param y: a PETSc vector to store the result
        :type y: Optional[PETSc.Vec], default is None

        :rtype: PETSc.Vec
        """

    @raise_not_implemented_error
    def apply_hermitian_transpose_mat(
        self: "LinearOperator", X: SLEPc.BV, Y: Optional[SLEPc.BV] = None
    ) -> SLEPc.BV:
        r"""
        Compute :math:`Y = L^*X`

        :param X: a SLEPc BV
        :type X: SLEPc.BV
        :param Y: a SLEPc BV to store the result
        :type Y: Optional[SLEPc.BV], default is None

        :rtype: SLEPc.BV
        """

    @raise_not_implemented_error
    def solve(
        self: "LinearOperator", x: PETSc.Vec, y: Optional[PETSc.Vec] = None
    ) -> PETSc.Vec:
        r"""
        Compute :math:`y = L^{-1}x`.

        :param x: a PETSc vector
        :type x: PETSc.Vec
        :param y: a PETSc vector to store the result
        :type y: Optional[PETSc.Vec], default is None

        :rtype: PETSc.Vec
        """

    @raise_not_implemented_error
    def solve_mat(
        self: "LinearOperator", X: SLEPc.BV, Y: Optional[SLEPc.BV] = None
    ) -> SLEPc.BV:
        r"""
        Compute :math:`Y = L^{-1}X`

        :param X: a SLEPc BV
        :type X: SLEPc.BV
        :param Y: a SLEPc BV to store the result
        :type Y: Optional[SLEPc.BV], default is None

        :rtype: SLEPc.BV
        """

    @raise_not_implemented_error
    def solve_hermitian_transpose(
        self: "LinearOperator", x: PETSc.Vec, y: Optional[PETSc.Vec] = None
    ) -> PETSc.Vec:
        r"""
        Compute :math:`y = L^{-*}x`.

        :param x: a PETSc vector
        :type x: PETSc.Vec
        :param y: a PETSc vector to store the result
        :type y: Optional[PETSc.Vec], default is None

        :rtype: PETSc.Vec
        """

    @raise_not_implemented_error
    def solve_hermitian_transpose_mat(
        self: "LinearOperator", X: SLEPc.BV, Y: Optional[SLEPc.BV] = None
    ) -> SLEPc.BV:
        r"""
        Compute :math:`Y = L^{-*}X`

        :param X: a SLEPc BV
        :type X: SLEPc.BV
        :param Y: a SLEPc BV to store the result
        :type Y: Optional[SLEPc.BV], default is None

        :rtype: SLEPc.BV
        """



__all__ = [
    "compute_gramian_factors",
    "compute_balanced_projection",
    "assemble_reduced_order_tensors",
]

import typing

import numpy as np
from mpi4py import MPI
from petsc4py import PETSc
from slepc4py import SLEPc

from ..utils.comms import compute_local_size
from ..utils.matrix import create_dense_matrix
from ..linear_operators import LinearOperator


def compute_gramian_factors(
    L_generators: typing.Callable,
    frequencies: np.ndarray,
    weights: np.ndarray,
    B: SLEPc.BV,
    C: SLEPc.BV,
) -> typing.Tuple[PETSc.Mat, PETSc.Mat]:
    r"""
    Compute the Gramian factors as outlined in section 2 of [Dergham2011]_.
    In particular, we approximate the Reachability and Observability
    Gramians as follows

    .. math::

        \begin{align}
        G_R &= \frac{1}{2\pi}\int_{-\infty}^{\infty}R(\omega)BB^*R(\omega)^*\
            \,d\omega \approx \sum_j w_j X(\omega_j)X(\omega_j)^* = XX^* \\
        G_O &= \frac{1}{2\pi}\int_{-\infty}^{\infty}R(\omega)^*CC^*R(\omega)\,\
            d\omega \approx \sum_j w_j Y(\omega_j)Y(\omega_j)^* = YY^*
        \end{align}
        
    where :math:`\omega_j\in\Omega` are quadrature points and :math:`w_j` are
    corresponding quadrature weights. Here, :math:`X(\omega_j) = R(\omega_j)B`,
    and :math:`R(\omega) = \left(i\omega I - A\right)^{-1}`.
    When the linear operator :math:`A` is real valued, the Gramians may be 
    further manipulated into

    .. math::

        G_R = \frac{1}{\pi}\int_{0}^{\infty}R(\omega)BB^*R(\omega)^*\
        \,d\omega \approx \sum_j \delta_j w_j\left(X_\mathrm{r}(\omega_i)\
        X_\mathrm{r}(\omega)^* + X_{\mathrm{i}}(\omega_j)\
            X_{\mathrm{i}}(\omega_j)^*\right),

    where the subscripts "r" and "i" denote the real and imaginary parts,
    and :math:`\delta_j = 1` if :math:`\omega_j = 0` and 
    :math:`\delta_j = 1/2` otherwise.

    :param L_generators: tuple of functions that define the linear operator \
        :math:`R(\omega)^{-1}`, the action of :math:`R(\omega)` on matrices,
        and destructors to avoid memory leaks. (See examples.)
    :type L_generators: Tuple[[Callable, Callable, Tuple[Callable, Callable, \
        ...]]]
    :param frequencies: quadrature points :math:`\omega_j`
    :type frequencies: numpy.ndarray
    :param weights: quadrature weights :math:`w_j` (see description above to 
        see how the definition of :math:`w_j` changes depending on whether the 
        system is real-valued).
    :type weights: numpy.ndarray
    :param B: input matrix
    :type B: SLEPc.BV
    :param C: output matrix
    :type C: SLEPc.BV

    :return: tuple :math:`(X, Y)` (see definitions above)
    :rtype: Tuple[PETSc.Mat, PETSc.Mat]

    References
    ----------
    .. [Dergham2011] Dergham et al., *Model reduction for fluids using \
        frequential snapshots*, Physics of Fluids, 2011
    """
    nb = B.getSizes()[-1]
    nc = C.getSizes()[-1]
    idces = np.argsort(frequencies).reshape(-1)
    frequencies = frequencies[idces]
    weights = weights[idces]
    min_f = frequencies[0]
    if min_f >= 0.0:
        split = True
        nf = (
            2 * (len(frequencies) - 1) + 1
            if min_f == 0.0
            else 2 * len(frequencies)
        )
    else:
        split = False
        nf = len(frequencies)

    L, action, destroyers = L_generators[0](frequencies[0])
    action_ht = (
        L.apply_hermitian_transpose_mat
        if action == L.apply_mat
        else L.solve_hermitian_transpose_mat
    )

    LB = L.create_left_bv(nb)
    LC = L.create_right_bv(nc)
    Xarray = np.zeros((LB.getSizes()[0][0], nb * nf), dtype=np.complex128)
    Yarray = np.zeros((LC.getSizes()[0][0], nc * nf), dtype=np.complex128)
    for k in range(len(frequencies)):
        if k > 0:
            L, action, destroyers = L_generators[k](frequencies[k])
            action_ht = (
                L.apply_hermitian_transpose_mat
                if action == L.apply_mat
                else L.solve_hermitian_transpose_mat
            )
        LB = action(B, LB)
        LC = action_ht(C, LC)
        LB.scale(np.sqrt(weights[k]))
        LC.scale(np.sqrt(weights[k]))
        LBMat = LB.getMat()
        LBMat_ = LBMat.getDenseArray().copy()
        LB.restoreMat(LBMat)
        LCMat = LC.getMat()
        LCMat_ = LCMat.getDenseArray().copy()
        LC.restoreMat(LCMat)
        if not split:  # The system is complex-valued
            Xarray[:, k * nb : (k + 1) * nb] = LBMat_
            Yarray[:, k * nc : (k + 1) * nc] = LCMat_
        else:  # The system is real-valued
            delta = 1.0
            if k == 0:
                # Handle the zero frequency separately if necessary
                delta = np.sqrt(1 / 2) if frequencies[k] == 0.0 else delta
                Xarray[:, 0:nb] = delta * LBMat_.real
                Yarray[:, 0:nc] = delta * LCMat_.real
            else:
                k0 = nb + 2 * (k - 1) * nb
                k1 = k0 + 2 * nb
                Xarray[:, k0:k1] = delta * np.concatenate(
                    (LBMat_.real, LBMat_.imag), -1
                )
                k0 = nc + 2 * (k - 1) * nc
                k1 = k0 + 2 * nc
                Yarray[:, k0:k1] = delta * np.concatenate(
                    (LCMat_.real, LCMat_.imag), -1
                )
        for destroy in destroyers:
            destroy()
    Xsizes = (LB.getSizes()[0], Xarray.shape[-1])
    Ysizes = (LC.getSizes()[0], Yarray.shape[-1])
    X = PETSc.Mat().createDense(Xsizes, None, Xarray, PETSc.COMM_WORLD)
    Y = PETSc.Mat().createDense(Ysizes, None, Yarray, PETSc.COMM_WORLD)
    LB.destroy()
    LC.destroy()
    return (X, Y)


def compute_balanced_projection(
    X: SLEPc.BV, Y: SLEPc.BV, r: int
) -> typing.Tuple[SLEPc.BV, SLEPc.BV, np.ndarray]:
    r"""
    Given the output :math:`(X, Y)` of :func:`.compute_gramian_factors`, compute
    :math:`\Phi` and :math:`\Psi` (each of dimension :math:`N\times r`).
    Given the singular value decomposition :math:`Y^*X = U\Sigma V^*`, these
    are given by

    .. math::

        \Phi = X V \Sigma^{-1/2},\quad \Psi = Y U \Sigma^{-1/2}.

    :param X: reachability Gramian factor
    :type X: PETSc.Mat
    :param Y: observability Gramian factor
    :type Y: PETSc. Mat
    :param r: number of columns of :math:`\Phi` and :math:`\Psi`
    :type r: int

    :return: tuple :math:`(\Phi, \Psi, \Sigma)`
    :rtype: Tuple[SLEPc.BV, SLEPc.BV, numpy.ndarray]
    """
    comm = PETSc.COMM_WORLD
    # Compute product Y^*@X
    Y.hermitianTranspose()
    Z = Y.matMult(X)
    Y.hermitianTranspose()
    svd = SLEPc.SVD().create(comm)
    svd.setOperators(Z)
    svd.setProblemType(SLEPc.SVD.ProblemType.STANDARD)
    svd.setType(SLEPc.SVD.Type.CROSS)
    svd.setWhichSingularTriplets(SLEPc.SVD.Which.LARGEST)
    svd.setUp()
    svd.solve()
    # Extract singular triplets
    r = np.min([r, svd.getConverged()])
    rloc = compute_local_size(r)
    u = Z.createVecLeft()
    v = Z.createVecRight()
    U = create_dense_matrix(comm, (Y.getSizes()[-1], (rloc, r)))
    V = create_dense_matrix(comm, (X.getSizes()[-1], (rloc, r)))
    S = create_dense_matrix(comm, ((rloc, r), (rloc, r)))
    U_ = U.getDenseArray()
    V_ = V.getDenseArray()
    S_ = np.zeros((r, r), dtype=np.float64)
    for i in range(r):
        s = svd.getSingularTriplet(i, u, v)
        U_[:, i] = u.getArray().copy()
        V_[:, i] = v.getArray().copy()
        S_[i, i] = s
        S.setValue(i, i, 1.0 / np.sqrt(s), addv=False)
        # Rotate the modes to minimize the imaginary part (this will keep
        # the modes real if the underlying system is real-valued)
        angle = 0
        if comm.Get_rank() == 0:
            idx = np.argmax(np.abs(U_[:, i].imag))
            angle = np.angle(U_[idx, i])
        angle = comm.tompi4py().bcast(angle, root=0)
        U_[:, i] *= np.exp(-1j * angle)
        V_[:, i] *= np.exp(-1j * angle)

    S.assemble(None)
    u.destroy()
    v.destroy()
    svd.destroy()
    # Compute matrices Phi and Psi for projection
    VS = V.matMult(S, None)
    US = U.matMult(S, None)
    Phi_mat = X.matMult(VS, None)
    Psi_mat = Y.matMult(US, None)
    Phi_ = SLEPc.BV().createFromMat(Phi_mat)
    Phi_.setType("mat")
    Phi = Phi_.copy()
    Psi_ = SLEPc.BV().createFromMat(Psi_mat)
    Psi_.setType("mat")
    Psi = Psi_.copy()
    objs = [Phi_, Psi_, Phi_mat, Psi_mat, U, V, S, Z, VS, US]
    for obj in objs:
        obj.destroy()
    return (Phi, Psi, S_)


def assemble_reduced_order_tensors(
    L: LinearOperator,
    B: SLEPc.BV,
    C: SLEPc.BV,
    Phi: SLEPc.BV,
    Psi: SLEPc.BV,
) -> typing.Tuple[np.ndarray, np.ndarray, np.ndarray]:
    r"""
    Assemble the reduced-order tensors :math:`A_r = \Psi^\intercal L \Phi`,
    :math:`B_r = \Psi^\intercal B` and :math:`C_r = \Phi^\intercal C`.

    :param L: linear operator governing the dynamics of the full-order system
    :type L: LinearOperator
    :param B: input matrix of the full-order system
    :type B: SLEPc.BV
    :param C: (complex-conj. transpose) of the output matrix of the full-order
        system
    :type C: SLEPc.BV

    :return: Tuple with :math:`A_r`, :math:`B_r` and :math:`C_r`
    :rtype: Tuple[numpy.ndarray, numpy.ndarray, numpy.ndarray]
    """
    LPhi = L.apply_mat(Phi)
    Amat = LPhi.dot(Psi)
    Bmat = B.dot(Psi)
    Cmat = C.dot(Phi)
    Ar = Amat.getDenseArray().copy()
    Br = Bmat.getDenseArray().copy()
    Cr = Cmat.getDenseArray().copy()
    objs = [LPhi, Amat, Bmat, Cmat]
    for obj in objs:
        obj.destroy()
    return Ar, Br, Cr



__all__ = [
    "read_vector",
    "read_coo_matrix",
    "read_harmonic_balanced_matrix",
    "read_dense_matrix",
    "read_bv",
    "read_harmonic_balanced_bv",
    "read_harmonic_balanced_vector",
    "write_to_file",
]

import typing

import numpy as np
from petsc4py import PETSc
from slepc4py import SLEPc

from .bv import bv_conj
from .comms import compute_local_size
from .matrix import convert_coo_to_csr


def read_vector(
    filename: str,
    sizes: typing.Optional[typing.Tuple[int, int]] = None,
) -> PETSc.Vec:
    r"""
    Read PETSc vector from file

    :param filename: name of the file that holds the vector
    :type filename: str
    :param sizes: :code:`(local size, global size)`
    :type sizes: Optional[Tuple[int, int]]

    :rtype: PETSc.Vec
    """
    comm = PETSc.COMM_WORLD
    viewer = PETSc.Viewer().createMPIIO(filename, "r", comm=comm)
    vec = PETSc.Vec().create(comm)
    vec.setSizes(sizes) if sizes != None else None
    vec.load(viewer)
    viewer.destroy()
    return vec


def read_coo_matrix(
    filenames: typing.Tuple[str, str, str],
    sizes: typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]],
) -> PETSc.Mat:
    r"""
    Read COO matrix from file

    :param filenames: names of the files that hold the rows, columns and values
        of the sparse matrix (e.g., :code:`(rows, cols, values)`)
    :type filenames: Tuple[str, str, str]
    :param sizes: :code:`((local rows, global rows), (local cols, global cols))`
    :type sizes: Tuple[Tuple[int, int], Tuple[int, int]]

    :rtype: PETSc.Mat
    """
    # Read COO vectors
    fname_rows, fname_cols, fname_vals = filenames
    rowsvec = read_vector(fname_rows)
    colsvec = read_vector(fname_cols)
    valsvec = read_vector(fname_vals)
    rows = np.asarray(rowsvec.getArray().real, dtype=PETSc.IntType)
    cols = np.asarray(colsvec.getArray().real, dtype=PETSc.IntType)
    vals = valsvec.getArray()
    # Delete zeros for efficiency
    idces = np.argwhere(np.abs(vals) <= 1e-16)
    rows = np.delete(rows, idces)
    cols = np.delete(cols, idces)
    vals = np.delete(vals, idces)
    # Convert COO to CSR and create the sparse matrix
    rows_ptr, cols, vals = convert_coo_to_csr([rows, cols, vals], sizes)
    M = PETSc.Mat().createAIJ(sizes, comm=PETSc.COMM_WORLD)
    M.setPreallocationCSR((rows_ptr, cols))
    M.setValuesCSR(rows_ptr, cols, vals, True)
    M.assemble(False)
    rowsvec.destroy()
    colsvec.destroy()
    valsvec.destroy()
    return M


def read_harmonic_balanced_matrix(
    filenames_lst: typing.List[typing.Tuple[str, str, str]],
    real_bflow: bool,
    block_sizes: typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]],
    full_sizes: typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]],
) -> PETSc.Mat:
    r"""
    Given :math:`\{\ldots, A_{-1}, A_{0}, A_{1},\ldots\}`, where :math:`A_j` is
    the :math:`j` th Fourier coefficient of a time-periodic matrix :math:`A(t)`,
    assemble the harmonic-balanced matrix

    .. math::

        M = \begin{bmatrix}
        \ddots & \ddots & \ddots & \ddots \\
        \ddots & A_{0} & A_{-1} & A_{-2} & \ddots \\
        \ddots & A_{1} & A_{0} & A_{-1} & A_{-2} & \ddots \\
        \ddots & A_{2} & A_{1} & A_{0} & A_{-1} & A_{-2} &\ddots \\
        & \ddots & A_{2} & A_{1} & A_{0} & A_{-1} &\ddots \\
        &  & \ddots & A_{2} & A_{1} & A_{0} & \ddots\\
        & & & \ddots & \ddots & \ddots & \ddots \\
        \end{bmatrix}
    

    :param filenames_lst: list of tuples, with each tuple of the form
        :code:`(rows_j.dat, cols_j.dat, vals_j.dat)` containing the COO arrays
        of the matrix :math:`A_j`
    :type filenames_lst: List[Tuple[str, str, str]]
    :param real_bflow: set to :code:`True` if :code:`filenames_lst` 
        contains the names of the COO arrays of the positive Fourier 
        coefficients of :math:`A(t)` (i.e., :math:`\{A_0, A_1, A_2, \ldots\}`). 
        The negative frequencies are assumed the complex-conjugates of the 
        positive ones. Set to :code:`False` otherwise
        (i.e., :math:`\{\ldots, A_{-2}, A_{-1}, A_0, A_1, A_2, \ldots\}`).
    :type real_bflow: bool
    :param block_sizes: sizes :code:`((local rows, global rows), 
        (local cols, global cols))` of :math:`A_j`
    :type block_sizes: Tuple[Tuple[int, int], Tuple[int, int]]
    :param full_sizes: sizes :code:`((local rows, global rows), 
        (local cols, global cols))` of :math:`M`
    :type full_sizes: Tuple[Tuple[int, int], Tuple[int, int]]

    :return: Sparse PETSc matrix :math:`M`
    :rtype: PETSc.Mat
    """
    # Read list of COO vectors
    rows_lst, cols_lst, vals_lst = [], [], []
    for filenames in filenames_lst:
        fname_rows, fname_cols, fname_vals = filenames
        rowsvec = read_vector(fname_rows)
        colsvec = read_vector(fname_cols)
        valsvec = read_vector(fname_vals)
        rowsvec_arr = np.asarray(
            rowsvec.getArray().real, dtype=PETSc.IntType
        ).copy()
        colsvec_arr = np.asarray(
            colsvec.getArray().real, dtype=PETSc.IntType
        ).copy()
        valsvec_arr = valsvec.getArray().copy()
        rows_lst.append(rowsvec_arr)
        cols_lst.append(colsvec_arr)
        vals_lst.append(valsvec_arr)
        rowsvec.destroy()
        colsvec.destroy()
        valsvec.destroy()

    if real_bflow:
        l = len(rows_lst)
        for i in range(1, l):
            idx_lst = i - l
            rows_lst.insert(0, rows_lst[idx_lst])
            cols_lst.insert(0, cols_lst[idx_lst])
            vals_lst.insert(0, np.conj(vals_lst[idx_lst]))

    rows, cols, vals = [], [], []
    Nrb = block_sizes[0][-1]  # Number of rows for each block
    Ncb = block_sizes[-1][-1]  # Number of columns for each block
    nblocks = full_sizes[0][-1] // Nrb  # Number of blocks
    nfb = (len(rows_lst) - 1) // 2  # Number of baseflow frequencies
    nfp = (nblocks - 1) // 2  # Number of perturbation frequencies
    if nfp < nfb:
        raise ValueError(
            f"The number of blocks must be larger than the number of Fourier "
            f"coefficients of A(t). (See function description.)"
        )
    for i in range(2 * nfp + 1):
        for j in range(2 * nfp + 1):
            k = i - j + nfb
            if k >= 0 and k < (2 * nfb + 1):
                rows.extend(rows_lst[k] + i * Nrb)
                cols.extend(cols_lst[k] + j * Ncb)
                vals.extend(vals_lst[k])
    rows = np.asarray(rows, dtype=PETSc.IntType)
    cols = np.asarray(cols, dtype=PETSc.IntType)
    vals = np.asarray(vals)

    # Delete zeros for efficiency
    idces = np.argwhere(np.abs(vals) <= 1e-16)
    rows = np.delete(rows, idces)
    cols = np.delete(cols, idces)
    vals = np.delete(vals, idces)
    # Convert COO to CSR and create the sparse matrix
    rows_ptr, cols, vals = convert_coo_to_csr([rows, cols, vals], full_sizes)
    M = PETSc.Mat().createAIJ(full_sizes, comm=PETSc.COMM_WORLD)
    M.setPreallocationCSR((rows_ptr, cols))
    M.setValuesCSR(rows_ptr, cols, vals, True)
    M.assemble(False)

    return M


def read_dense_matrix(
    filename: str,
    sizes: typing.Tuple[typing.Tuple[int, int], typing.Tuple[int, int]],
) -> PETSc.Mat:
    r"""
    Read dense PETSc matrix from file.

    :param filename: name of the file that holds the matrix
    :type filename: str
    :param sizes: :code:`((local rows, global rows), (local cols, global cols))`
    :type sizes: Tuple[Tuple[int, int], Tuple[int, int]]

    :rtype: PETSc.Mat
    """
    comm = PETSc.COMM_WORLD
    viewer = PETSc.Viewer().createMPIIO(filename, "r", comm=comm)
    M = PETSc.Mat().createDense(sizes, comm=comm)
    M.load(viewer)
    viewer.destroy()
    return M


def read_bv(
    filename: str,
    sizes: typing.Tuple[typing.Tuple[int, int], int],
) -> SLEPc.BV:
    r"""
    Read dense matrix from file and store as a SLEPc BV

    :param filename: name of the file that holds the matrix
    :type filename: str
    :param sizes: :code:`((local rows, global rows), global columns)`
    :type sizes: Tuple[Tuple[int, int], int]

    :rtype: SLEPc.BV
    """
    ncols = sizes[-1]
    sizes_mat = (sizes[0], (compute_local_size(ncols), ncols))
    Mm = read_dense_matrix(filename, sizes_mat)
    M = SLEPc.BV().createFromMat(Mm)
    M.setType("mat")
    Mm.destroy()
    return M


def read_harmonic_balanced_bv(
    filenames_lst: typing.List[str],
    real_bflow: bool,
    block_sizes: typing.Tuple[typing.Tuple[int, int], int],
    full_sizes: typing.Tuple[typing.Tuple[int, int], int],
) -> SLEPc.BV:
    r"""
    Given :math:`\{\ldots, A_{-1}, A_{0}, A_{1},\ldots\}`, where :math:`A_j` is
    the :math:`j` th Fourier coefficient of a time-periodic matrix :math:`A(t)`,
    assemble the harmonic-balanced matrix

    .. math::

        M = \begin{bmatrix}
        \ddots & \ddots & \ddots & \ddots \\
        \ddots & A_{0} & A_{-1} & A_{-2} & \ddots \\
        \ddots & A_{1} & A_{0} & A_{-1} & A_{-2} & \ddots \\
        \ddots & A_{2} & A_{1} & A_{0} & A_{-1} & A_{-2} &\ddots \\
        & \ddots & A_{2} & A_{1} & A_{0} & A_{-1} &\ddots \\
        &  & \ddots & A_{2} & A_{1} & A_{0} & \ddots\\
        & & & \ddots & \ddots & \ddots & \ddots \\
        \end{bmatrix}
    
    :param comm: MPI communicator (PETSc.COMM_WORLD)
    :type comm: PETSc.Comm
    :param filenames_lst: list of names of files where the Fourier modes 
        :math:`A_j` are stored as SLEPc BVs
    :type filenames_lst: List[str]
    :param real_bflow: set to :code:`True` if :code:`filenames_lst` 
        contains the names of the COO arrays of the positive Fourier 
        coefficients of :math:`A(t)` (i.e., :math:`\{A_0, A_1, A_2, \ldots\}`). 
        The negative frequencies are assumed the complex-conjugates of the 
        positive ones. Set to :code:`False` otherwise
        (i.e., :math:`\{\ldots, A_{-2}, A_{-1}, A_0, A_1, A_2, \ldots\}`).
    :type real_bflow: bool
    :param block_sizes: sizes :code:`((local rows, global rows), global_cols)` 
        of :math:`A_j`
    :type block_sizes: Tuple[Tuple[int, int], int]
    :param full_sizes: sizes :code:`((local rows, global rows), global_cols)` 
        of :math:`M`
    :type full_sizes: Tuple[Tuple[int, int], int]

    :return: SLEPc BV :math:`M`
    :rtype: SLEPc.BV
    """
    # Read list of BVs
    bvs_lst = []
    for filename in filenames_lst:
        bv = read_bv(filename, block_sizes)
        bvs_lst.append(bv.copy())
        bv.destroy()

    if real_bflow:
        l = len(bvs_lst)
        for i in range(1, l):
            idx_lst = i - l
            bvs_lst.insert(0, bv_conj(bvs_lst[idx_lst]))

    Nrb = block_sizes[0][-1]  # Number of rows for each block
    Ncb = block_sizes[-1]  # Number of cols for each block
    nblocks = full_sizes[0][-1] // Nrb  # Number of blocks
    nfb = (len(bvs_lst) - 1) // 2  # Number of baseflow frequencies
    nfp = (nblocks - 1) // 2  # Number of perturbation frequencies
    if nfp < nfb:
        raise ValueError(
            f"The number of blocks must be larger than the number of Fourier "
            f"coefficients of A(t). (See function description.)"
        )

    bv_mat = bvs_lst[0].getMat()
    r0, r1 = bv_mat.getOwnershipRange()
    bvs_lst[0].restoreMat(bv_mat)
    rows = np.arange(r0, r1, dtype=PETSc.IntType)
    cols = np.arange(Ncb, dtype=PETSc.IntType)
    M = SLEPc.BV().create(PETSc.COMM_WORLD)
    M.setSizes(full_sizes[0], full_sizes[-1])
    M.setType("mat")
    Mmat = M.getMat()
    for i in range(2 * nfp + 1):
        for j in range(2 * nfp + 1):
            k = i - j + nfb
            if k >= 0 and k < (2 * nfb + 1):
                rows_k = rows + i * Nrb
                cols_k = cols + j * Ncb
                bv_k_mat = bvs_lst[k].getMat()
                array = bv_k_mat.getDenseArray().reshape(-1)
                Mmat.setValues(rows_k, cols_k, array, None)
                bvs_lst[k].restoreMat(bv_k_mat)
    Mmat.assemble(None)
    M.restoreMat(Mmat)
    for bv in bvs_lst:
        bv.destroy()
    return M


def read_harmonic_balanced_vector(
    filenames_lst: typing.List[str],
    real_bflow: bool,
    block_sizes: typing.Tuple[int, int],
    full_sizes: typing.Tuple[int, int],
) -> PETSc.Vec:
    r"""
    Given :math:`\{\ldots, v_{-1}, v_{0}, v_{1},\ldots\}`, where :math:`v_j` is
    the :math:`j` th Fourier coefficient of a time-periodic vector :math:`v(t)`,
    assemble

    .. math::

        \hat{v} = \begin{bmatrix}
            \vdots \\ v_{-1} \\ v_0 \\ v_{1} \\ \vdots
        \end{bmatrix}.

    :param filenames_lst: list of names of files where the Fourier modes
        :math:`v_j` are stored as PETSc Vec
    :type filenames_lst: List[str]
    :param real_bflow: set to :code:`True` if :code:`filenames_lst`
        contains the names of the COO arrays of the positive Fourier
        coefficients of :math:`v(t)` (i.e., :math:`\{v_0, v_1, v_2, \ldots\}`).
        The negative frequencies are assumed the complex-conjugates of the
        positive ones. Set to :code:`False` otherwise
        (i.e., :math:`\{\ldots, v_{-2}, v_{-1}, v_0, v_1, v_2, \ldots\}`).
    :type real_bflow: bool
    :param block_sizes: sizes :code:`(local rows, global rows)` of :math:`v_j`
    :type block_sizes: Tuple[int, int]
    :param full_sizes: sizes :code:`(local rows, global rows)` of :math:`\hat{v}`
    :type full_sizes: Tuple[int, int]

    :return: vector :math:`\hat{v}`
    :rtype: PETSc.Vec
    """
    vec_lst = []
    for filename in filenames_lst:
        vec = read_vector(filename, block_sizes)
        vec_lst.append(vec.copy())
        vec.destroy()

    if real_bflow:
        l = len(vec_lst)
        for i in range(1, l):
            idx_lst = i - l
            vecconj = vec_lst[idx_lst].copy()
            vecconj.conjugate()
            vec_lst.insert(0, vecconj.copy())
            vecconj.destroy()

    Vec = PETSc.Vec().create(comm=PETSc.COMM_WORLD)
    Vec.setSizes(full_sizes)
    Vec.setUp()
    r0, _ = vec_lst[0].getOwnershipRange()
    nfb = (len(vec_lst) - 1) // 2  # Number of baseflow frequencies
    nfp = (
        full_sizes[-1] // block_sizes[-1] - 1
    ) // 2  # Number of perturbation frequencies
    for i in range(2 * nfb + 1):
        j = i + (nfp - nfb)
        rows = (
            j * block_sizes[-1]
            + np.arange(block_sizes[0], dtype=PETSc.IntType)
            + r0
        )
        Vec.setValues(rows, vec_lst[i].getArray(), False)
    Vec.assemble()

    for vec in vec_lst:
        vec.destroy()

    return Vec


def write_to_file(
    filename: str,
    object: typing.Union[PETSc.Mat, PETSc.Vec, SLEPc.BV],
) -> None:
    r"""
    Write PETSc/SLEPc object (PETSc.Mat, PETSc.Vec or SLEPc.BV) to file.

    :param filename: name of the file to store the object
    :type filename: str
    :param object: any PETSc matrix or vector, or SLEPc BV
    :type object: Union[PETSc.Mat, PETSc.Vec, SLEPc.BV]
    """
    viewer = PETSc.Viewer().createBinary(filename, "w", comm=object.getComm())
    if isinstance(object, SLEPc.BV):
        mat = object.getMat()
        mat.view(viewer)
        object.restoreMat(mat)
    else:
        object.view(viewer)
    viewer.destroy()



__all__ = [
    "create_mumps_solver",
    "check_lu_factorization",
    "create_gmres_bjacobi_solver",
    "check_gmres_bjacobi_solver",
]

import typing

from petsc4py import PETSc

from .miscellaneous import petscprint
from .random import generate_random_petsc_vector


def create_mumps_solver(A: PETSc.Mat) -> PETSc.KSP:
    r"""
    Compute an LU factorization of the matrix A using MUMPS.

    :param A: PETSc matrix
    :type A: PETSc.Mat

    :return ksp: PETSc KSP solver
    :rtype ksp: PETSc.KSP
    """
    ksp = PETSc.KSP().create(comm=A.getComm())
    ksp.setOperators(A)
    ksp.setType("preonly")
    ksp.setUp()
    pc = ksp.getPC()
    pc.setType("lu")
    pc.setFactorSolverType("mumps")
    pc.setReusePreconditioner(True)
    pc.setUp()
    return ksp


def check_lu_factorization(A: PETSc.Mat, ksp: PETSc.KSP) -> None:
    r"""
    Check that the LU factorization computed in :func:`.create_mumps_solver`
    has succeeded.

    :param A: PETSc matrix
    :type A: PETSc.Mat
    :param ksp: PETSc KSP solver
    :type ksp: PETSc.KSP
    """
    sizes = A.getSizes()[0]
    b = generate_random_petsc_vector(sizes)
    x = b.duplicate()
    ksp.solve(b, x)
    pc = ksp.getPC()
    Mat = pc.getFactorMatrix()
    Infog1 = Mat.getMumpsInfog(1)
    Infog2 = Mat.getMumpsInfog(2)
    if Infog1 != 0:
        raise ValueError(
            f"MUMPS factorization failed with INFO(1) = {Infog1}  "
            f"and INFO(2) = {Infog2}'"
        )
    x.destroy()
    b.destroy()


def create_gmres_bjacobi_solver(
    A: PETSc.Mat,
    nblocks: int,
    rtol: typing.Optional[float] = 1e-10,
    atol: typing.Optional[float] = 1e-10,
    monitor: typing.Optional[bool] = False,
) -> PETSc.KSP:
    r"""
    Create GMRES solver with block-jacobi preconditioner.

    :param A: PETSc matrix
    :type A: PETSc.Mat
    :param nblocks: number of blocks for the block jacobi preconditioner
    :type nblocks: int
    :param rtol: relative tolerance for GMRES
    :type rtol: Optional[float], default is :math:`10^{-10}`
    :param atol: absolute tolerance for GMRES
    :type atol: Optional[float], default is :math:`10^{-10}`
    :param monitor: :code:`True` to monitor convergence and print residual
        history to terminal. :code:`False` otherwise
    :type monitor: Optional[bool], default is :code:`False`

    :return ksp: PETSc KSP solver
    :rtype ksp: PETSc.KSP
    """

    comm = A.getComm()
    monitor_fun = None
    if monitor:

        def monitor_fun(ksp, its, rnorm):
            string = f"GMRES Iteration {its:3d}, Residual Norm = {rnorm:.3e}"
            petscprint(comm, string)

    opts = PETSc.Options()
    opts["pc_type"] = "bjacobi"
    opts["pc_bjacobi_blocks"] = nblocks
    opts["sub_ksp_type"] = "preonly"
    opts["sub_pc_type"] = "lu"
    opts["sub_pc_factor_mat_solver_type"] = "mumps"

    ksp = PETSc.KSP().create(comm=comm)
    ksp.setOperators(A)
    ksp.setType("gmres")
    ksp.setTolerances(rtol=rtol, atol=atol)
    ksp.setMonitor(monitor_fun)
    pc = ksp.getPC()
    pc.setFromOptions()
    pc.setUp()
    ksp.setUp()

    return ksp


def check_gmres_bjacobi_solver(A: PETSc.Mat, ksp: PETSc.KSP) -> None:
    r"""
    Check that the solver computed in :func:`.create_gmres_bjacobi_solver`
    has succeeded.

    :param A: PETSc matrix
    :type A: PETSc.Mat
    :param ksp: PETSc KSP solver
    :type ksp: PETSc.KSP
    """
    sizes = A.getSizes()[0]
    b = generate_random_petsc_vector(sizes)
    x = b.duplicate()
    ksp.solve(b, x)
    x.destroy()
    b.destroy()
    reason = ksp.getConvergedReason()
    if reason < 0:
        raise ValueError(
            f"GMRES solver did not converge. ConvergedReason = {reason}"
        )



__all__ = [
    "enforce_complex_conjugacy",
    "check_complex_conjugacy",
    "vec_real",
    "vec_imag",
]

import numpy as np
import typing
from petsc4py import PETSc


def vec_real(
    x: PETSc.Vec, inplace: typing.Optional[bool] = False
) -> PETSc.Vec:
    r"""
    Returns the real part :math:`\text{Re}(x)` of the PETSc Vec.

    :type x: PETSc.Vec
    :param inplace: in-place if :code:`True`, else the result is stored in a
        new PETSc.Vec
    :type inplace: Optional[bool], default is False

    :rtype: PETSc.Vec
    """
    y = x if inplace else x.copy()
    ya = y.getArray()
    ya[:] = ya.real
    return y


def vec_imag(
    x: PETSc.Vec, inplace: typing.Optional[bool] = False
) -> PETSc.Vec:
    r"""
    Returns the imaginary part :math:`\text{Im}(x)` of the PETSc Vec.

    :type x: PETSc.Vec
    :param inplace: in-place if :code:`True`, else the result is stored in a
        new PETSc.Vec
    :type inplace: Optional[bool], default is False

    :rtype: PETSc.Vec
    """
    y = x if inplace else x.copy()
    ya = y.getArray()
    ya[:] = ya.imag
    return y


def enforce_complex_conjugacy(
    comm: PETSc.Comm, vec: PETSc.Vec, nblocks: int
) -> None:
    r"""
    Suppose we have a vector

    .. math::
        v = \left(\ldots,v_{-1},v_{0},v_{1},\ldots\right)

    where :math:`v_i` are complex vectors. This function enforces
    :math:`v_{-i} = \overline{v_{i}}` for all :math:`i` (this implies that
    :math:`v_0` will be purely real).

    :param vec: vector :math:`v` described above
    :type vec: PETSc.Vec
    :param nblocks: number of vectors :math:`v_i` in :math:`v`. This must
        be an odd number.
    :type nblocks: int

    :rtype: None
    """
    if np.mod(nblocks, 2) == 0:
        raise ValueError(
            "The number of blocks must be an odd number. "
            "Currently you set {nblocks} blocks."
        )
    scatter, vec_seq = PETSc.Scatter().toZero(vec)
    scatter.begin(vec, vec_seq, addv=PETSc.InsertMode.INSERT)
    scatter.end(vec, vec_seq, addv=PETSc.InsertMode.INSERT)
    if comm.getRank() == 0:
        array = vec_seq.getArray()
        block_size = len(array) // nblocks
        for i in range(nblocks // 2):
            j = nblocks - 1 - i
            i0, i1 = i * block_size, (i + 1) * block_size
            j0, j1 = j * block_size, (j + 1) * block_size
            array[i0:i1] = array[j0:j1].conj()
        i = nblocks // 2
        i0, i1 = i * block_size, (i + 1) * block_size
        array[i0:i1] = array[i0:i1].real
        vec_seq.setValues(np.arange(len(array), dtype=PETSc.IntType), array)
        vec_seq.assemble()
    scatter.begin(
        vec_seq,
        vec,
        addv=PETSc.InsertMode.INSERT,
        mode=PETSc.ScatterMode.REVERSE,
    )
    scatter.end(
        vec_seq,
        vec,
        addv=PETSc.InsertMode.INSERT,
        mode=PETSc.ScatterMode.REVERSE,
    )
    scatter.destroy()
    vec_seq.destroy()


def check_complex_conjugacy(
    comm: PETSc.Comm, vec: PETSc.Vec, nblocks: int
) -> bool:
    r"""
    Verify whether the components :math:`v_i` of the vector

    .. math::
        v = \left(\ldots,v_{-1},v_{0},v_{1},\ldots\right)

    satisfy :math:`v_{-i} = \overline{v_{i}}` for all :math:`i`.

    :param vec: vector :math:`v` described above
    :type vec: PETSc.Vec
    :param nblocks: number of vectors :math:`v_i` in :math:`v`. This must
        be an odd number.
    :type nblocks: int

    :return: :code:`True` if the components are complex-conjugates of each
        other and :code:`False` otherwise
    :rtype: Bool
    """
    if np.mod(nblocks, 2) == 0:
        raise ValueError(
            "The number of blocks must be an odd number. "
            "Currently you set {nblocks} blocks."
        )
    scatter, vec_seq = PETSc.Scatter().toZero(vec)
    scatter.begin(vec, vec_seq, addv=PETSc.InsertMode.INSERT)
    scatter.end(vec, vec_seq, addv=PETSc.InsertMode.INSERT)
    cc = None
    if comm.getRank() == 0:
        array = vec_seq.getArray()
        block_size = len(array) // nblocks
        array_block = np.zeros(block_size, dtype=np.complex128)
        for i in range(nblocks):
            i0, i1 = i * block_size, (i + 1) * block_size
            array_block += array[i0:i1]
        cc = True if np.linalg.norm(array_block.imag) <= 1e-13 else False
    scatter.destroy()
    vec_seq.destroy()
    cc = comm.tompi4py().bcast(cc, root=0)
    return cc


def assemble_harmonic_balanced_vector(
    vec_lst: typing.List[PETSc.Vec],
    bflow_freqs: np.array,
    pertb_freqs: np.array,
    sizes: typing.Tuple[int, int],
) -> PETSc.Vec:
    if len(bflow_freqs) != len(vec_lst):
        raise ValueError(
            f"Error in assemble_harmonic_balanced_vector(). vec_lst "
            f"should have the same length as bflow_freqs."
        )

    put_back = False
    if np.min(bflow_freqs) == 0.0:
        put_back = True
        for i in range(1, len(bflow_freqs)):
            idx_lst = i - 1 - nfp
            vecconj = vec_lst[idx_lst].copy()
            vecconj.conjugate()
            vec_lst.insert(0, vecconj)
        bflow_freqs = np.concatenate(
            (-np.flipud(bflow_freqs[1:]), bflow_freqs)
        )

    # Create the harmonic-balanced BV
    Vec = PETSc.Vec().create(comm=PETSc.COMM_WORLD)
    Vec.setSizes(sizes)
    Vec.setUp()
    r0, _ = Vec.getOwnershipRange()
    nfb = (len(bflow_freqs) - 1) // 2  # Number of baseflow frequencies
    nfp = (len(pertb_freqs) - 1) // 2  # Number of perturbation frequencies
    vec_sizes = vec_lst[0].getSizes()
    nrows_loc, nrows = vec_sizes[0]
    for i in range(2 * nfb + 1):
        j = i + (nfp - nfb)
        rows = j * nrows + np.arange(nrows_loc, dtype=PETSc.IntType) + r0
        Vec.setValues(rows, vec_lst[j].getArray(), False)
    Vec.assemble()

    if put_back:
        bflow_freqs = bflow_freqs[nfb:]
        for i in range(nfb):
            vec_lst[i].destroy()
        vec_lst[nfb:]

    return Vec



__all__ = ["get_mpi_type", "petscprint", "get_memory_usage"]

import tracemalloc

import numpy as np
from mpi4py import MPI
from petsc4py import PETSc

numpy_to_mpi_dtype = {
    np.dtype(np.int32): MPI.INT,
    np.dtype(np.int64): MPI.INT64_T,
    np.dtype(np.float64): MPI.DOUBLE,
    np.dtype(np.complex64): MPI.COMPLEX,
    np.dtype(np.complex128): MPI.DOUBLE_COMPLEX,
}


def get_mpi_type(dtype: np.dtype) -> MPI.Datatype:
    r"""
    Get the corresponding MPI type for a given numpy data type.

    :param dtype: (e.g., :code:`np.dtype(PETSc.IntType)`)
    :type dtype: np.dtypes
    :rtype: MPI.Datatype
    """
    mpi_dtype = numpy_to_mpi_dtype.get(dtype)
    if mpi_dtype is None:
        raise ValueError(f"No MPI type found for numpy dtype {dtype}")
    return mpi_dtype


def petscprint(comm: PETSc.Comm, arg: any) -> None:
    r"""
    Print to terminal

    :param comm: MPI communicator (PETSc.COMM_WORLD or PETSc.COMM_SELF)
    :type comm: PETSc.Comm
    :param arg: argument to be fed into print()
    :type arg: any
    """
    if comm == PETSc.COMM_SELF or comm == MPI.COMM_SELF:
        print(arg)
    else:
        if PETSc.COMM_WORLD.getRank() == 0:
            print(arg)


def get_memory_usage(comm: PETSc.Comm) -> float:
    r"""
    Compute the used memory (in Mb) across the MPI pool

    :type comm: PETSc.Comm
    :rtype: float
    """
    comm = comm.tompi4py()
    snapshot = tracemalloc.take_snapshot()
    top_stats = snapshot.statistics("lineno")
    total_memory = sum(stat.size for stat in top_stats) / (1024**2)
    value = (
        sum(comm.allgather(total_memory))
        if comm == MPI.COMM_WORLD
        else total_memory
    )
    return value
